"""I3 device base classes."""

import asyncio
from collections import namedtuple
from typing import Dict, List, Tuple, Union

from ..config import (
    ACK_A_SCENE,
    BLUE_LED_OFF,
    BRIGHTNESS_MAX,
    BRIGHTNESS_MIN,
    BRIGHTNESS_START,
    CLEANUP_REPORT_OFF,
    DETACH_LOAD_ON,
    DISABLE_BUTTONS,
    DO_NOT_ROTATE_TO_OFF,
    GREEN_LED_OFF,
    INSTEON_OFF,
    KEY_BEEP_ON,
    LED_BLINK_ON_ERROR_ON,
    LED_BLINK_ON_TX_ON,
    LED_BRIGHTNESS,
    LED_OFF,
    NIGHT_MODE_LED_BRIGHTNESS,
    NIGHT_MODE_MAX_LEVEL,
    NIGHT_MODE_ON,
    NIGHT_MODE_RAMP_RATE,
    NO_CACHE,
    NON_TOGGLE_MASK,
    NON_TOGGLE_ON_OFF_MASK,
    OFF_MASK,
    ON_LEVEL,
    ON_MASK,
    OPERATING_FLAGS,
    PROGRAM_LOCK_ON,
    RAMP_RATE,
    RED_LED_OFF,
    RELAY_AT_FULL_ON,
    RELAY_MODE_OFF,
    RESUME_DIM_ON,
    RF_DISABLE_ON,
    SKIP_SOME_HOPS,
    TEND_ON,
    UNUSED,
    YAKETY_YAK,
)
from ..config.extended_property import ExtendedProperty
from ..config.op_flag_property_byte import OpFlagPropertyByte
from ..config.operating_flag import OperatingFlag
from ..constants import PropertyType, ResponseStatus
from ..groups.on_level import OnLevel
from ..handlers.from_device.manual_change import ManualChangeInbound
from ..handlers.from_device.off_at_ramp_rate import OffAtRampRateInbound
from ..handlers.from_device.on_at_ramp_rate import OnAtRampRateInbound
from ..handlers.to_device.factory_reset import FactoryResetCommand
from ..handlers.to_device.night_mode_off import NightModeOffCommand
from ..handlers.to_device.night_mode_on import NightModeOnCommand
from ..managers.ext_prop_read_manager import ExtendedPropertyReadManager
from ..managers.ext_prop_write_manager import ExtendedPropertyWriteManager
from ..utils import multiple_status
from .device_commands import MANUAL_CHANGE, OFF_AT_RAMP_RATE, ON_AT_RAMP_RATE
from .on_off_responder_base import OnOffResponderBase
from .variable_responder_base import VariableResponderBase

OP_FLAG_1F_00 = f"{OPERATING_FLAGS}_1f_00"
OP_FLAG_1F_05 = f"{OPERATING_FLAGS}_1f_05"
OP_FLAG_DATA_4 = f"{OPERATING_FLAGS}_data_4"

READ_MGRS = "ext_prop_read_mgrs"
WRITE_MGRS = "ext_prop_write_mgrs"
NIGHT_MODE_OFF = "night_mode_off"

OpsFlagDef = namedtuple(
    "OpsFlagDef", ["name", "group", "bit", "set_cmd", "unset_cmd", "prop_type"]
)
ExtPropDef = namedtuple(
    "ExtPropDef", ["name", "value_type", "is_read_only", "prop_type", "op_flags"]
)
ExtPropReadDef = namedtuple("ExtPropReadDef", ["cmd2", "data1", "data2", "data3"])
ExtPropRespDef = namedtuple(
    "ExtPropRespDef", ["cmd2", "data1", "data2", "data3", "props_dict"]
)
ExtPropWriteDef = namedtuple(
    "ExtPropWriteDef", ["cmd2", "data1", "data2", "props_dict"]
)


##############################################################################
# This section defines the default operating flags, extened properties and
# extened properties readers and writers.
##############################################################################
def default_i3_ops_flags(
    dimmable: bool, additional_flags: List[OpsFlagDef]
) -> List[OpsFlagDef]:
    """Return a list of default operating flags for i3 devices."""
    show = PropertyType.STANDARD if dimmable else PropertyType.HIDDEN
    def_flags = [
        # Group 0
        OpsFlagDef(PROGRAM_LOCK_ON, 0, 0, 0x00, 0x01, PropertyType.STANDARD),
        OpsFlagDef(LED_BLINK_ON_TX_ON, 0, 1, 0x02, 0x03, PropertyType.STANDARD),
        OpsFlagDef(RESUME_DIM_ON, 0, 2, 0x04, 0x05, show),
        OpsFlagDef(f"{UNUSED}_0_3", 0, 3, None, None, PropertyType.HIDDEN),
        OpsFlagDef(LED_OFF, 0, 4, 0x08, 0x09, PropertyType.STANDARD),
        OpsFlagDef(KEY_BEEP_ON, 0, 5, 0x0A, 0x0B, PropertyType.STANDARD),
        OpsFlagDef(RF_DISABLE_ON, 0, 6, 0x0C, 0x0D, PropertyType.ADVANCED),
        # SHOULD NEVER USE IT WILL MAKE THE DEVICE UNRESPONSIVE
        # However, it can be set with 20 0F if you really want to
        OpsFlagDef(INSTEON_OFF, 0, 7, None, None, PropertyType.HIDDEN),
        # Group 5
        OpsFlagDef(TEND_ON, 5, 0, None, None, PropertyType.HIDDEN),
        # -- Night mode is set with (cmd1: 0x3B  cmd2: 0xXX) and (cmd2: 0x3C  cmd2: 0xXX)
        OpsFlagDef(NIGHT_MODE_ON, 5, 1, None, None, PropertyType.STANDARD),
        OpsFlagDef(LED_BLINK_ON_ERROR_ON, 5, 2, None, None, PropertyType.STANDARD),
        OpsFlagDef(CLEANUP_REPORT_OFF, 5, 3, None, None, PropertyType.HIDDEN),
        OpsFlagDef(DISABLE_BUTTONS, 5, 4, None, None, PropertyType.HIDDEN),
        # -- Unsetting Detached load does not work with the documented command 20 1B
        # -- So setting as read_only but can be set via OP_FLAG_1F_05
        OpsFlagDef(DETACH_LOAD_ON, 5, 5, 0x1B, 0x1A, PropertyType.HIDDEN),
        OpsFlagDef(ACK_A_SCENE, 5, 6, 0x1C, 0x1D, PropertyType.HIDDEN),
        OpsFlagDef(RELAY_MODE_OFF, 5, 7, 0x1E, 0x1F, show),
        # Group 7
        OpsFlagDef(YAKETY_YAK, 7, 0, None, None, PropertyType.HIDDEN),
        OpsFlagDef(RED_LED_OFF, 7, 1, 0x28, 0x29, PropertyType.STANDARD),
        OpsFlagDef(SKIP_SOME_HOPS, 7, 2, 0x2A, 0x2B, PropertyType.HIDDEN),
        OpsFlagDef(GREEN_LED_OFF, 7, 3, 0x2C, 0x2D, PropertyType.STANDARD),
        OpsFlagDef(BLUE_LED_OFF, 7, 4, 0x2E, 0x2F, PropertyType.STANDARD),
        OpsFlagDef(NO_CACHE, 7, 5, 0x30, 0x31, PropertyType.HIDDEN),
        OpsFlagDef(DO_NOT_ROTATE_TO_OFF, 7, 6, 0x32, 0x33, show),
        OpsFlagDef(RELAY_AT_FULL_ON, 7, 7, None, None, PropertyType.HIDDEN),
        # Group 9 (KPL Only?)
        OpsFlagDef(f"{NIGHT_MODE_ON}_kpl", 9, 0, None, None, PropertyType.HIDDEN),
        OpsFlagDef(f"{NO_CACHE}_kpl", 9, 1, None, None, PropertyType.HIDDEN),
    ]
    # Create an index of the group and bit of the additional flags
    add_flag_set: List[Tuple[int, int]] = []
    for add_flag in additional_flags:
        add_flag_set.append((add_flag.group, add_flag.bit))

    # Remove duplicates keeping the additional flag
    flags = [flag for flag in def_flags if (flag.group, flag.bit) not in add_flag_set]
    flags.extend(additional_flags)
    return flags


def operating_flag_ext_prop_def(
    group: int,
    operating_flags: List[OperatingFlag],
    op_flags_def: List[OpsFlagDef],
) -> Dict[int, OperatingFlag]:
    """Return a dictionary of operating flags for use in an Extended Property byte."""
    op_flag_byte_def = get_op_flags_by_group(group=group, op_flags=op_flags_def)
    return {index: operating_flags[name] for index, name in op_flag_byte_def.items()}


def default_i3_ext_props(
    dimmable,
    operating_flags: List[OperatingFlag],
    op_flags_def: List[OpsFlagDef],
    op_flags_data_4: Dict[int, str],
    buttons: Union[List[int], None] = None,
) -> List[ExtPropDef]:
    """Return a list of default extended properties for i3 devices."""
    show_dim = PropertyType.STANDARD if dimmable else PropertyType.HIDDEN
    show_dim_adv = PropertyType.ADVANCED if dimmable else PropertyType.HIDDEN

    op_flag_byte_00_dict = operating_flag_ext_prop_def(0, operating_flags, op_flags_def)
    op_flag_byte_05_dict = operating_flag_ext_prop_def(5, operating_flags, op_flags_def)
    op_flag_byte_data_4_dict = {
        bit: operating_flags[name] for bit, name in op_flags_data_4.items()
    }

    props = [
        ExtPropDef(NIGHT_MODE_LED_BRIGHTNESS, int, False, PropertyType.STANDARD, None),
        ExtPropDef(NIGHT_MODE_RAMP_RATE, int, False, PropertyType.STANDARD, None),
        ExtPropDef(LED_BRIGHTNESS, int, False, PropertyType.STANDARD, None),
        ExtPropDef(BRIGHTNESS_MIN, int, False, show_dim, None),
        ExtPropDef(BRIGHTNESS_MAX, int, False, show_dim, None),
        ExtPropDef(BRIGHTNESS_START, int, False, show_dim, None),
        ExtPropDef(NIGHT_MODE_MAX_LEVEL, int, False, PropertyType.STANDARD, None),
        ExtPropDef(
            OP_FLAG_1F_00,
            None,
            None,
            PropertyType.HIDDEN,
            op_flags=op_flag_byte_00_dict,
        ),
        ExtPropDef(
            OP_FLAG_1F_05,
            None,
            None,
            PropertyType.HIDDEN,
            op_flags=op_flag_byte_05_dict,
        ),
        ExtPropDef(
            OP_FLAG_DATA_4,
            int,
            False,
            PropertyType.HIDDEN,
            op_flags=op_flag_byte_data_4_dict,
        ),
    ]
    if buttons:
        for button in buttons:
            if button == 1:
                on_mask = ON_MASK
                off_mask = OFF_MASK
                ramp_rate = RAMP_RATE
                on_level = ON_LEVEL
            else:
                on_mask = f"{ON_MASK}_{button}"
                off_mask = f"{OFF_MASK}_{button}"
                ramp_rate = f"{RAMP_RATE}_{button}"
                on_level = f"{ON_LEVEL}_{button}"
            props.append(ExtPropDef(on_mask, int, False, PropertyType.ADVANCED, None))
            props.append(ExtPropDef(off_mask, int, False, PropertyType.ADVANCED, None))
            props.append(ExtPropDef(ramp_rate, int, False, PropertyType.ADVANCED, None))
            props.append(ExtPropDef(on_level, int, False, PropertyType.STANDARD, None))

        props.append(
            ExtPropDef(NON_TOGGLE_MASK, int, False, PropertyType.ADVANCED, None)
        )
        props.append(
            ExtPropDef(NON_TOGGLE_ON_OFF_MASK, int, False, PropertyType.ADVANCED, None)
        )
    else:
        props.append(ExtPropDef(RAMP_RATE, int, False, show_dim_adv, None))
        props.append(ExtPropDef(ON_LEVEL, int, False, show_dim, None))

    return props


def properties_2e_00_xx_00_def(
    properties: Dict[str, ExtendedProperty], groups: List[int]
) -> List[Tuple[ExtPropReadDef, ExtPropRespDef, List[ExtPropWriteDef]]]:
    """Return a set of properties that represent the 2E 00 XX 00 properties."""
    if 0 not in groups and 1 not in groups:
        raise ValueError(
            "Properties for 2E 00 XX 00 reader/writes requires either buttons 0 or 1."
        )
    prop_group_read_resp_def: List[Tuple[ExtPropReadDef, ExtPropRespDef]] = []
    for group in groups:
        props: Dict[int, ExtendedProperty] = {}
        writers: List[ExtPropWriteDef] = []
        if group in [0, 1]:
            props[7] = properties[RAMP_RATE]
            props[8] = properties[ON_LEVEL]
            props[9] = properties[LED_BRIGHTNESS]
            writers.append(
                ExtPropWriteDef(0x00, group, 0x05, {3: properties[RAMP_RATE]})
            )
            writers.append(
                ExtPropWriteDef(0x00, group, 0x06, {3: properties[ON_LEVEL]})
            )
            writers.append(
                ExtPropWriteDef(0x00, 0x00, 0x07, {3: properties[LED_BRIGHTNESS]})
            )
        if group == 1:
            props[3] = properties[ON_MASK]
            props[4] = properties[OFF_MASK]
            props[10] = properties[NON_TOGGLE_MASK]
            props[13] = properties[NON_TOGGLE_ON_OFF_MASK]
            writers.append(ExtPropWriteDef(0x00, group, 0x02, {3: properties[ON_MASK]}))
            writers.append(
                ExtPropWriteDef(0x00, group, 0x03, {3: properties[OFF_MASK]})
            )
            writers.append(
                ExtPropWriteDef(0x00, 0x00, 0x08, {3: properties[NON_TOGGLE_MASK]})
            )
            writers.append(
                ExtPropWriteDef(
                    0x00, 0x00, 0x0B, {3: properties[NON_TOGGLE_ON_OFF_MASK]}
                )
            )
        elif group not in [0, 1]:
            props[3] = properties[f"{ON_MASK}_{group}"]
            props[4] = properties[f"{OFF_MASK}_{group}"]
            props[7] = properties[f"{RAMP_RATE}_{group}"]
            props[8] = properties[f"{ON_LEVEL}_{group}"]
            writers.append(
                ExtPropWriteDef(
                    0x00, group, 0x02, {3: properties[f"{ON_MASK}_{group}"]}
                )
            )
            writers.append(
                ExtPropWriteDef(
                    0x00, group, 0x03, {3: properties[f"{OFF_MASK}_{group}"]}
                )
            )
            writers.append(
                ExtPropWriteDef(
                    0x00, group, 0x05, {3: properties[f"{RAMP_RATE}_{group}"]}
                )
            )
            writers.append(
                ExtPropWriteDef(
                    0x00, group, 0x06, {3: properties[f"{ON_LEVEL}_{group}"]}
                )
            )
        read_def = ExtPropReadDef(0x00, group, 0x00, 0x00)
        resp_def = ExtPropRespDef(0x00, group, 0x01, None, props)
        prop_group_read_resp_def.append((read_def, resp_def, writers))
    return prop_group_read_resp_def


def properties_2e_01_reader_writer_def(
    properties: Dict[str, ExtendedProperty], resp_data_1: int
) -> Tuple[ExtPropReadDef, ExtPropRespDef, ExtPropWriteDef]:
    """Return a set of properties that represent the 2E 01 properties."""
    props_dict = {
        2: properties[OP_FLAG_1F_00],
        3: properties[OP_FLAG_1F_05],
        4: properties[OP_FLAG_DATA_4],
        5: properties[RAMP_RATE],
        6: properties[ON_LEVEL],
        7: properties[LED_BRIGHTNESS],
        8: properties[BRIGHTNESS_MAX],
        9: properties[BRIGHTNESS_MIN],
        10: properties[BRIGHTNESS_START],
        11: properties[NIGHT_MODE_MAX_LEVEL],
        12: properties[NIGHT_MODE_LED_BRIGHTNESS],
        13: properties[NIGHT_MODE_RAMP_RATE],
    }
    read_def = ExtPropReadDef(0x01, 0x00, 0x00, 0x00)
    resp_def = ExtPropRespDef(0x01, resp_data_1, None, None, props_dict)
    writer_def = ExtPropWriteDef(0x01, 0x02, None, props_dict)
    return read_def, resp_def, writer_def


def get_op_flags_by_group(group: int, op_flags: List[OpsFlagDef]) -> Dict[int, str]:
    """Return a dictionary of bit and operating flag bit definition by group."""
    return {flag.bit: flag.name for flag in op_flags if flag.group == group}


##############################################################################
# End of default Operating Flags and Extended Properties definition section
##############################################################################


# pylint: disable=no-member
# pylint: disable=super-with-arguments
class I3Base:
    """I3 device type base class."""

    _is_kpl: bool = False
    _op_flags_data_4: Dict[int, str] = {
        0: YAKETY_YAK,
        1: RED_LED_OFF,
        2: SKIP_SOME_HOPS,
        3: GREEN_LED_OFF,
        4: BLUE_LED_OFF,
        5: NO_CACHE,
        6: DO_NOT_ROTATE_TO_OFF,
        7: RELAY_AT_FULL_ON,
    }  # used for 2E 01 read and write command

    def __init__(
        self, address, cat, subcat, firmware=0x00, description="", model="", **kwargs
    ):
        """Init the DeviceBattery class."""

        # The super class is a Device class such as VariableResponderBase
        super(I3Base, self).__init__(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            description=description,
            model=model,
            **kwargs,
        )

    async def async_read_ext_properties(self, group=None) -> ResponseStatus:
        """Get the device extended properties."""
        responses = []
        for mgr in self._managers[READ_MGRS]:
            response = await mgr.async_send()
            responses.append(response)
            await asyncio.sleep(0.2)
        return multiple_status(*responses)

    async def async_write_op_flags(self) -> ResponseStatus:
        """Write the operating flags to the device."""
        responses = []
        if prop := self._operating_flags.get(NIGHT_MODE_ON):
            if prop.is_dirty:
                response = await self.async_set_night_mode(prop.new_value)
                responses.append(response)
        response = await self._op_flags_manager.async_write()
        responses.append(response)
        return multiple_status(*responses)

    async def async_write_ext_properties(self) -> ResponseStatus:
        """Write the extended properties."""
        responses = [ResponseStatus.SUCCESS]
        for cmd in self._managers[WRITE_MGRS]:
            if cmd.is_dirty:
                responses.append(await cmd.async_write())
        return multiple_status(*responses)

    async def async_set_night_mode(self, on: bool) -> ResponseStatus:
        """Set the device night mode on or off."""
        cmd = NIGHT_MODE_ON if on else NIGHT_MODE_OFF
        for _ in range(0, 3):
            response = await self._managers[cmd].async_send()
            await asyncio.sleep(0.5)
        return response

    async def async_factory_reset(self) -> ResponseStatus:
        """Reset this device to factory defaults."""
        cmd = FactoryResetCommand(self._address, self._cat, self._subcat)
        return await cmd.async_send()

    def _register_handlers_and_managers(self) -> None:
        super(I3Base, self)._register_handlers_and_managers()
        self._managers[NIGHT_MODE_ON] = NightModeOnCommand(self._address)
        self._managers[NIGHT_MODE_OFF] = NightModeOffCommand(self._address)
        groups = list(self._groups) if self._is_kpl else [0]
        reader_writer_groups = properties_2e_00_xx_00_def(self._properties, groups)
        for prop_reader, prop_resp, prop_writers in reader_writer_groups:
            self._add_ext_prop_read_manager(
                properties=prop_resp.props_dict,
                cmd2=prop_reader.cmd2,
                data1_read=prop_reader.data1,
                data2_read=prop_reader.data2,
                data3_read=prop_reader.data3,
                data1_resp=prop_resp.data1,
                data2_resp=prop_resp.data2,
                data3_resp=prop_resp.data3,
            )
            for prop_writer in prop_writers:
                self._add_ext_prop_write_manager(
                    properties=prop_writer.props_dict,
                    cmd2=prop_writer.cmd2,
                    data1=prop_writer.data1,
                    data2=prop_writer.data2,
                )

        # Cat   Subcat  Model   Device      Data1
        # 0x01  0x57    PS01    i3 Paddle   0x05
        # 0x01	0x58    DS01    i3 Dial     0x04
        # 0x01	0x59    KP014   i3 Keypad   0x01
        # 0x02	0x3F    WR01    i3 Outlet   0x03
        data2_map = {
            (0x01, 0x57): 0x05,
            (0x01, 0x58): 0x04,
            (0x01, 0x59): 0x01,
            (0x02, 0x3F): 0x03,
        }
        data_1_resp = data2_map.get((self._cat, self._subcat))
        if not data_1_resp:
            raise ValueError("Cannot find data 1 value for the 2E 01 message response.")
        prop_reader, prop_resp, prop_writer = properties_2e_01_reader_writer_def(
            self._properties, data_1_resp
        )
        self._add_ext_prop_read_manager(
            properties=prop_resp.props_dict,
            cmd2=prop_reader.cmd2,
            data1_read=prop_reader.data1,
            data2_read=prop_reader.data2,
            data3_read=prop_reader.data3,
            data1_resp=prop_resp.data1,
            data2_resp=prop_resp.data2,
            data3_resp=prop_resp.data3,
        )
        self._add_ext_prop_write_manager(
            properties=prop_writer.props_dict,
            cmd2=prop_writer.cmd2,
            data1=prop_writer.data1,
            data2=prop_writer.data2,
        )

    def _subscribe_to_handelers_and_managers(self) -> None:
        super(I3Base, self)._subscribe_to_handelers_and_managers()
        self._managers[NIGHT_MODE_ON].subscribe(self._handle_night_mode)
        self._managers[NIGHT_MODE_OFF].subscribe(self._handle_night_mode)

    def _handle_night_mode(self, night_mode) -> None:
        """Handle the night mode on/off response."""
        if prop := self._operating_flags.get(NIGHT_MODE_ON):
            prop.set_value(night_mode)

    def _register_default_op_flags_and_props(
        self,
        dimmable: bool,
        additional_flags: List[OpsFlagDef],
    ) -> None:
        # Adding operating flags
        flags = default_i3_ops_flags(dimmable, additional_flags)
        for flag in flags:
            self._add_operating_flag(
                name=flag.name,
                group=flag.group,
                bit=flag.bit,
                set_cmd=flag.set_cmd,
                unset_cmd=flag.unset_cmd,
                prop_type=flag.prop_type,
            )
        # Need to set Night Mode as read/write because it is not updated with a 0x20 command
        self._operating_flags[NIGHT_MODE_ON].is_read_only = False
        flag_names = [
            LED_BLINK_ON_ERROR_ON,
            CLEANUP_REPORT_OFF,
            DISABLE_BUTTONS,
            NIGHT_MODE_ON,
            DETACH_LOAD_ON,
            DO_NOT_ROTATE_TO_OFF,
            RELAY_AT_FULL_ON,
        ]
        for name in flag_names:
            flag = self._operating_flags.get(name)
            if flag:
                flag.is_read_only = False

        # Adding extended properties
        buttons = list(self._groups) if self._is_kpl else None
        for prop in default_i3_ext_props(
            dimmable, self._operating_flags, flags, self._op_flags_data_4, buttons
        ):
            if prop.op_flags:
                self._properties[prop.name] = OpFlagPropertyByte(
                    address=self._address, name=prop.name, flags=prop.op_flags
                )
            else:
                self._properties[prop.name] = ExtendedProperty(
                    address=self._address,
                    name=prop.name,
                    value_type=prop.value_type,
                    is_reversed=False,
                    is_read_only=prop.is_read_only,
                    prop_type=prop.prop_type,
                )

    def _add_ext_prop_read_manager(
        self,
        properties: Dict[int, ExtendedProperty],
        cmd2: int = 0,
        data1_read: int = 0,
        data2_read: int = 0,
        data3_read: int = None,
        data1_resp: int = 0,
        data2_resp: int = 1,
        data3_resp: int = None,
    ) -> None:
        self._managers[READ_MGRS] = self._managers.get(READ_MGRS, [])
        self._managers[READ_MGRS].append(
            ExtendedPropertyReadManager(
                address=self._address,
                properties=properties,
                cmd2=cmd2,
                data1_read=data1_read,
                data2_read=data2_read,
                data3_read=data3_read,
                data1_resp=data1_resp,
                data2_resp=data2_resp,
                data3_resp=data3_resp,
            )
        )

    def _add_ext_prop_write_manager(
        self,
        properties: Dict[int, ExtendedProperty],
        data2: Union[int, None],
        cmd2: int = 0,
        data1: int = 0,
    ) -> None:
        self._managers[WRITE_MGRS] = self._managers.get(WRITE_MGRS, [])
        self._managers[WRITE_MGRS].append(
            ExtendedPropertyWriteManager(
                address=self._address,
                properties=properties,
                cmd2=cmd2,
                data1=data1,
                data2=data2,
            )
        )


# pylint: disable=super-with-arguments
class I3VariableResponderBase(I3Base, VariableResponderBase):
    """I3 variable responder base class."""

    def _register_op_flags_and_props(self) -> None:
        self._register_default_op_flags_and_props(dimmable=True, additional_flags=[])

    def _register_handlers_and_managers(self) -> None:
        super()._register_handlers_and_managers()
        for group, group_prop in self._groups.items():
            if isinstance(group_prop, OnLevel):
                self._handlers[group][ON_AT_RAMP_RATE] = OnAtRampRateInbound(
                    self._address, group
                )
                self._handlers[group][OFF_AT_RAMP_RATE] = OffAtRampRateInbound(
                    self._address, group
                )
                self._handlers[group][MANUAL_CHANGE] = ManualChangeInbound(
                    self._address, group
                )

    def _subscribe_to_handelers_and_managers(self) -> None:
        super()._subscribe_to_handelers_and_managers()
        for group, group_prop in self._groups.items():
            if isinstance(group_prop, OnLevel):
                self._handlers[group][MANUAL_CHANGE].subscribe(self.async_status)
                # pylint: disable=cell-var-from-loop
                self._handlers[group][OFF_AT_RAMP_RATE].subscribe(
                    lambda on_level, ramp_rate: self._handle_on_off_at_ramp_rate(
                        group, on_level, ramp_rate
                    ),
                    force_strong_ref=True,
                )
                self._handlers[group][ON_AT_RAMP_RATE].subscribe(
                    lambda on_level, ramp_rate: self._handle_on_off_at_ramp_rate(
                        group, on_level, ramp_rate
                    ),
                    force_strong_ref=True,
                )

    def _handle_on_off_at_ramp_rate(self, group: int, on_level: int, ramp_rate: int):
        """Handle the on and off at ramp rate inbound broadcast messages."""
        self._groups[group].set_value(on_level)


# pylint: disable=super-with-arguments
class I3OnOffResponderBase(I3Base, OnOffResponderBase):
    """I3 on/off responder base class."""

    def _register_op_flags_and_props(self) -> None:
        self._register_default_op_flags_and_props(dimmable=False, additional_flags=[])



"""Modem command to get the IM configuration."""
import logging

from . import ack_handler, nak_handler
from ..topics import GET_IM_CONFIGURATION
from .outbound_base import OutboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class GetImConfigurationHandler(OutboundHandlerBase):
    """Handle Get IM Configuration commands."""

    def __init__(self):
        """Init the GetImConfigurationHandler class."""
        super().__init__(topic=GET_IM_CONFIGURATION)

    @ack_handler
    async def async_handle_ack(
        self, disable_auto_linking, monitor_mode, auto_led, deadman
    ):
        """Handle the ACK message from the modem."""
        await self._async_handle_ack()
        self._call_subscribers(
            disable_auto_linking=disable_auto_linking,
            monitor_mode=monitor_mode,
            auto_led=auto_led,
            deadman=deadman,
        )

    @nak_handler
    async def async_handle_nak(self):
        """Handle the NAK response from the modem."""
        await self._async_handle_nak()



"""Class to define the message definitions."""


def _calc_length(fields):
    """Calculate the length."""
    length = 2
    for field in fields:
        if isinstance(field, MessageDefinition):
            continue
        length += field.length
    return length


def _create_slices(fields):
    slices = []
    slices.append(slice(0, 1))
    slices.append(slice(1, 2))
    start = 2
    for field in fields:
        stop = start + field.length
        if field.length > 1:
            slices.append(slice(start, stop))
        else:
            slices.append(start)
        start = stop
    return slices


class MessageDefinition:
    """Insteon message defintion."""

    def __init__(self, message_id, message_fields: []):
        """Init the MessageDefinition class."""
        self._message_id = message_id
        self._fields = message_fields
        self._length = _calc_length(message_fields)
        self._slices = _create_slices(self.fields)

    def __len__(self):
        """Emit the message length."""
        return self._length

    @property
    def message_id(self):
        """Emit the message type."""
        return self._message_id

    @property
    def fields(self):
        """Emit the message fields."""
        return self._fields

    @property
    def slices(self):
        """Emit the message slices."""
        return self._slices



"""Manage identifying unknown devices."""
import asyncio
from binascii import unhexlify
from collections import namedtuple
import logging

import async_timeout

from .. import pub
from ..address import Address
from ..constants import DeviceAction, ResponseStatus
from ..handlers.all_link_completed import AllLinkCompletedHandler
from ..handlers.from_device.assign_to_all_link_group import AssignToAllLinkGroupCommand
from ..handlers.from_device.delete_from_all_link_group import (
    DeleteFromAllLinkGroupCommand,
)
from ..handlers.to_device.id_request import IdRequestCommand
from ..handlers.to_device.ping import PingCommand
from ..subscriber_base import SubscriberBase
from ..utils import subscribe_topic, unsubscribe_topic

_LOGGER = logging.getLogger(__name__)
MAX_RETRIES = 5
RETRY_PAUSE = 2
PING_DELAY = 20


DeviceId = namedtuple("DeviceId", "address cat subcat firmware")  # product_id')


def _normalize_identifier(value):
    """Convert a byte, str or int to int."""
    if value is None:
        return value
    if isinstance(value, (bytearray, bytes)):
        return int.from_bytes(value, "big")
    if isinstance(value, str):
        if value[0:2] == "0x":
            value = value[2:]
        return int.from_bytes(unhexlify(value), "big")
    return int(value)


class DeviceIdManager(SubscriberBase):
    """Manage device identities."""

    def __init__(self):
        """Init the DeviceIdManager class."""
        super().__init__(subscriber_topic="device_id")
        self._unknown_devices = []
        self._device_ids = {}

        # Cannot set queue here because we are outside the loop
        self._awake_devices_queue = None

        self._id_device_lock = asyncio.Lock()
        self._all_link_complete = AllLinkCompletedHandler()
        self._all_link_complete.subscribe(self._all_link_complete_received)
        self._ping_tasks = {}

    def __getitem__(self, address):
        """Return the unknown device list."""
        address = Address(address)
        return self._device_ids.get(address)

    def __setitem__(self, address, device_id):
        """Set a device ID."""
        self._device_ids[address] = device_id

    @property
    def unknown_devices(self):
        """Return a list of addresses where the device type has not been identified."""
        return self._unknown_devices

    def start(self):
        """Start the ID manager for unknown devices."""
        if self._awake_devices_queue is None:
            self._awake_devices_queue = asyncio.Queue()
        asyncio.ensure_future(self._id_awake_devices())

    def close(self):
        """Close the ID listener."""
        if self._awake_devices_queue is not None:
            self._awake_devices_queue.put_nowait(None)
        for _, task in self._ping_tasks.items():
            task.cancel()

    def append(self, address: Address, refresh=False):
        """Append a device address to the list."""
        address = Address(address)

        # If the address is already in the unknown device list stop
        if address in self._unknown_devices and not refresh:
            return

        # If it is not in the identified device list, add it as an unknown device
        device_id = self._device_ids.get(address)
        if refresh or device_id is None or device_id.cat is None:
            self._device_ids[address] = DeviceId(address, None, None, None)
            if address not in self._unknown_devices:
                self._unknown_devices.append(address)
            if address not in self._ping_tasks:
                task = asyncio.ensure_future(self._ping_device(address))
                self._ping_tasks[address] = task
            subscribe_topic(self._device_awake, address.id)

    def set_device_id(
        self,
        address: Address,
        cat: int,
        subcat: int,
        firmware: int = 0x00,
        link_mode=DeviceAction.ADDED,
    ):
        """Set the device ID of a device."""
        address = Address(address)
        cat = _normalize_identifier(cat)
        subcat = _normalize_identifier(subcat)
        firmware = _normalize_identifier(firmware)

        if address in self._unknown_devices:
            self._unknown_devices.remove(address)
        unsubscribe_topic(self._device_awake, address.id)
        if address in self._ping_tasks:
            self._ping_tasks[address].cancel()

        device_id = DeviceId(address, cat, subcat, firmware)
        self._device_ids[address] = device_id
        self._call_subscribers(device_id=device_id, link_mode=link_mode)

    async def async_id_devices(self, refresh: bool = False):
        """Identify the devices in the unknown device list."""
        if refresh:
            for address_id in self._device_ids:
                self.append(address_id, refresh=True)

        # We change the unknown device list during the process so we make a copy
        address_list = self._unknown_devices.copy()
        async with self._id_device_lock:
            for address in address_list:
                await self.async_id_device(address)
        return self._device_ids

    async def async_id_device(self, address: Address, refresh: bool = False):
        """Call ID Request command for all unknown devices."""

        received_queue = asyncio.Queue()

        async def async_device_id_received(device_id, link_mode):
            """Receive notification a device has been identified."""
            await received_queue.put(device_id)

        self.subscribe(async_device_id_received)

        id_handler = IdRequestCommand(address)
        id_response_handler = AssignToAllLinkGroupCommand(address)
        id_response_handler_alt = DeleteFromAllLinkGroupCommand(address)
        id_response_handler.subscribe(self._id_response)
        id_response_handler_alt.subscribe(self._id_response)

        address = Address(address)
        device_id = DeviceId(address, None, None, None)
        self.append(address, refresh)

        retries = MAX_RETRIES
        try:
            while retries:
                response = await id_handler.async_send()
                if response == ResponseStatus.SUCCESS:
                    try:
                        async with async_timeout.timeout(RETRY_PAUSE):
                            return await received_queue.get()
                    except asyncio.TimeoutError:
                        pass
                elif response in [
                    ResponseStatus.DIRECT_NAK_ALDB,
                    ResponseStatus.DIRECT_NAK_CHECK_SUM,
                    ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                ]:
                    _LOGGER.warning(
                        "Device %s refused the Device ID request with error: %s (%r)",
                        str(address),
                        str(response),
                        response,
                    )
                    return None

                retries -= 1
                await asyncio.sleep(RETRY_PAUSE)

        finally:
            id_response_handler.unsubscribe(self._id_response)
            id_response_handler_alt.unsubscribe(self._id_response)

        return device_id

    def _id_response(self, address, cat, subcat, firmware, group, link_mode):
        """Receive a device ID response."""
        self.set_device_id(
            address=address,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
            link_mode=link_mode,
        )

    async def _id_awake_devices(self):
        """Loop on devices that wake up and send a message."""
        if self._awake_devices_queue is None:
            self._awake_devices_queue = asyncio.Queue()
        while True:
            address = await self._awake_devices_queue.get()
            if address is None:
                break
            await asyncio.sleep(0.5)
            device_id = await self.async_id_device(address=address)
            if device_id and device_id.cat is None:
                subscribe_topic(self._device_awake, address.id)
            elif device_id is None:
                unsubscribe_topic(self._device_awake, address.id)
                if address in self._ping_tasks:
                    self._ping_tasks[address].cancel()
                break

    def _device_awake(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Identify an unknown device.

        An unknown device has sent a message so we try to identify it.

        """
        # If the async_id_devices method is running do not proceed
        if self._id_device_lock.locked():
            return

        try:
            address = Address(topic.name.split(".")[0])
        except ValueError:
            return

        unsubscribe_topic(self._device_awake, address.id)

        # Has the device been identified already
        device_id = self._device_ids.get(address)
        if device_id and device_id.cat is not None:
            return

        if self._awake_devices_queue is None:
            self._awake_devices_queue = asyncio.Queue()
        self._awake_devices_queue.put_nowait(address)

    def _all_link_complete_received(
        self, link_mode, group, target, cat, subcat, firmware
    ):
        """Receive All-Link complete message."""
        self._id_response(target, cat, subcat, firmware, group, link_mode)

    async def _ping_device(self, address):
        """Ping the device until awake."""
        cmd = PingCommand(address)
        response = None
        retry_wait = 60
        retries = (
            24 * 60 * 1.5
        )  # 1.5 days. If a devices does not respond in this time it is dead
        try:
            while retries and response != ResponseStatus.SUCCESS:
                device_id = self._device_ids.get(address)
                if device_id is not None and device_id.cat is not None:
                    return
                await asyncio.sleep(retry_wait)
                response = await cmd.async_send()
                if response in [
                    ResponseStatus.DIRECT_NAK_ALDB,
                    ResponseStatus.DIRECT_NAK_CHECK_SUM,
                    ResponseStatus.DIRECT_NAK_INVALID_COMMAND,
                    ResponseStatus.DIRECT_NAK_NO_LOAD,
                ]:
                    _LOGGER.warning(
                        "Device %s rejected an ID request with error: %s (%r)",
                        str(address),
                        response,
                        response,
                    )
                    return
                retries -= 1
        except asyncio.CancelledError:
            return



"""Manage creation of default links for a device."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ...constants import ResponseStatus

if TYPE_CHECKING:
    from ...device_types.device_base import Device

TIMEOUT = 3
_LOGGER = logging.getLogger(__name__)


async def async_add_default_links(device: Device):
    """Establish default links between the modem and device."""
    # pylint: disable=import-outside-toplevel
    from ... import devices

    if not device.aldb.is_loaded:
        await device.aldb.async_load()

    if not device.aldb.is_loaded:
        return ResponseStatus.UNSENT

    modem = devices.modem

    for link_info in device.default_links:
        is_controller = link_info.is_controller
        group = link_info.group
        device.aldb.add(
            group=group,
            target=modem.address,
            controller=is_controller,
            data1=int(modem.cat),
            data2=int(modem.subcat),
            data3=modem.firmware,
        )
        modem.aldb.add(
            group=group,
            target=device.address,
            controller=not is_controller,
            data1=int(device.cat),
            data2=int(device.subcat),
            data3=int(device.firmware),
        )
    _, failed_device = await device.aldb.async_write()
    _, failed_modem = await modem.aldb.async_write()
    if failed_modem == 0 and failed_device == 0:
        return ResponseStatus.SUCCESS
    return ResponseStatus.FAILURE



"""Manage outbound ON command to a device."""
from ...constants import EngineVersion
from ...topics import GET_INSTEON_ENGINE_VERSION
from .direct_command import DirectCommandHandlerBase


class EngineVersionRequest(DirectCommandHandlerBase):
    """Manage an outbound ON command to a device."""

    def __init__(self, address):
        """Init the OnLevelCommand class."""
        super().__init__(topic=GET_INSTEON_ENGINE_VERSION, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the OFF command async."""
        return await super().async_send()

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Handle the OFF response direct ACK."""
        try:
            version = EngineVersion(cmd2)
        except ValueError:
            # If Insteon creates a new Insteon Engine Version we want to record it as Other, not Unknown
            # Unknown is reserved for a device which has not had an engine version discovery performed
            version = EngineVersion.OTHER
        self._call_subscribers(engine_version=version)



"""Manage inbound ON command from device."""
import logging

from .. import inbound_handler
from ...constants import MessageFlagType
from ...topics import EXTENDED_GET_RESPONSE
from ..inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class ExtendedGetResponseHandler(InboundHandlerBase):
    """Off Level command inbound."""

    def __init__(self, address, cmd2=0x00, data1=None, data2=0x01, data3=None):
        """Init the OffInbound class."""
        super().__init__(
            topic=EXTENDED_GET_RESPONSE,
            address=address,
            message_type=MessageFlagType.DIRECT,
        )
        self._cmd2 = cmd2
        self._data1 = data1
        self._data2 = data2
        self._data3 = data3

        topic = f"handler.{self._address.id}.{EXTENDED_GET_RESPONSE}"
        for group in (cmd2, data1, data2, data3):
            if group is None:
                topic = f"{topic}.x"
            else:
                topic = f"{topic}.{group}"
        self._subscriber_topic = topic

    @inbound_handler
    def handle_response(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the Extended Get response from a device."""
        # pylint: disable=too-many-boolean-expressions
        if (
            cmd2 == self._cmd2
            and (self._data1 is None or user_data["d1"] == self._data1)
            and (self._data2 is None or user_data["d2"] == self._data2)
            and (self._data3 is None or user_data["d3"] == self._data3)
        ):
            data = {}
            for item in range(2, 15):
                data[f"data{item}"] = user_data[f"d{item}"]
            self._call_subscribers(group=user_data["d1"], data=data)



"""Mock message handling via Chain of Command pattern."""
from abc import ABCMeta
import asyncio
import logging

import async_timeout

from ..constants import ResponseStatus
from ..utils import build_topic, publish_topic
from .inbound_base import InboundHandlerBase

MSG_TIME = 180  # seconds to send each message in queue, used for timeout below
NAK_RETRIES = 9
NAK_RESEND_WAIT = 0.2
_LOGGER = logging.getLogger(__name__)


class OutboundHandlerBase(InboundHandlerBase):
    """Manage a message chain."""

    __meta__ = ABCMeta

    def __init__(self, topic, address=None, group=None, message_type=None):
        """Init the MessageManager."""
        self._message_response = asyncio.Queue()
        self._send_lock = asyncio.Lock()
        self._message_type = message_type
        super().__init__(topic, address=address, group=group, message_type=message_type)
        self._send_topic = build_topic(
            topic=self._topic,
            prefix="send",
            address=None,
            group=None,
            message_type=self._message_type,
        )
        self._nak_retries = NAK_RETRIES
        self._kwargs = {}

    @property
    def message_response(self) -> asyncio.Queue:
        """Message response queue to manage message status."""
        return self._message_response

    async def async_send(self, **kwargs):
        """Send the message and wait for a status."""
        async with self._send_lock:
            # Empty the message queue
            while not self._message_response.empty():
                try:
                    self._message_response.get_nowait()
                except asyncio.QueueEmpty:
                    pass

            self._nak_retries = NAK_RETRIES
            self._kwargs = kwargs
            publish_topic(self._send_topic, **self._kwargs)

            try:
                async with async_timeout.timeout(MSG_TIME):
                    return await self._message_response.get()
            except asyncio.TimeoutError:
                # Send a FAILURE message if this is running more than 180 seconds
                return ResponseStatus.FAILURE

    async def _async_handle_ack(self, **kwargs):
        """Handle the ACK processing."""
        await self._message_response.put(ResponseStatus.SUCCESS)

    async def _async_handle_nak(self, **kwargs):
        """Call from a command handler to handle the NAK response from the modem."""
        if not self._send_lock.locked():
            return
        if self._nak_retries:
            sleep_duration = (
                NAK_RESEND_WAIT * (NAK_RETRIES - self._nak_retries) + NAK_RESEND_WAIT
            )
            await asyncio.sleep(sleep_duration)
            publish_topic(self._send_topic, **self._kwargs)
            self._nak_retries -= 1
            return
        await self._message_response.put(ResponseStatus.FAILURE)



"""Base object for all command line menues."""

from argparse import ArgumentParser
import asyncio
from binascii import Error as BinasciiError, unhexlify
from cmd import Cmd
from collections import namedtuple
import getpass
import inspect
import logging
import os
import signal
import sys
import traceback
from typing import List

from .. import devices
from ..address import Address
from ..constants import RelayMode, ThermostatMode, ToggleMode
from ..x10_address import X10Address
from .log_filter import NoStdoutFilter, StdoutFilter, StripPrefixFilter
from .utils import patch_stdin_stdout, set_loop, stdio

_LOGGING = logging.getLogger(__name__)
LOG_FILE_NAME = "pyinsteon_tools.log"
STDOUT_LOG_HANDLER = "stdout_handler"
FILE_LOG_HANDLER = "file_handler"
CmdArgs = namedtuple("CmdArgs", "workdir device username host hub_version port")
ARG_TYPES = {
    "auto_led": bool,
    "awake_time": bytes,
    "button": range(1, 9),
    "buttons": [range(1, 9)],
    "deadman": bool,
    "disable_auto_linking": bool,
    "fast": bool,
    "force": bool,
    "group": bytes,
    "humidity": bytes,
    "link_mode": bool,
    "logging_mode": bool,
    "master": Address,
    "monitor_mode": bool,
    "on_level": bytes,
    "on": bool,
    "open_level": bytes,
    "relay_mode": RelayMode,
    "seconds": bytes,
    "temperature": bytes,
    "thermostat_mode": ThermostatMode,
    "toggle_mode": ToggleMode,
}
DEFAULT_HUB_PORT = {1: 9761, 2: 25105}


def validate_address(address, allow_all, match_device, log_stdout) -> List[Address]:
    """Check that an address value is valid.

    Returns `[address]` if `address` is a valid `Address` or if `address` eq `all`.

    Raises `ValueError` if:
        - `address` is not a valid `Address`
        - `address` eq `all` and `allow_all` is `False`
        - `address` does not match an existing device and `match_device` is `True`
    """
    if not address:
        raise ValueError("No address provided")

    if address != "all":
        try:
            address = Address(address)
        except ValueError:
            address = X10Address(address)
        if not match_device or address in devices:
            return [address]
        log_stdout(f"No device found with address {address}")
        raise ValueError(f"No device found with address {address}")

    if allow_all:
        return list(devices)
    log_stdout("A single address is expected")
    raise ValueError("A single address is expected")


class ToolsBase(Cmd):
    """Base class for all tools menues."""

    _active_task = None

    def __init__(self, loop, args=None, menu=None, stdin=None, stdout=None):
        """Init the InsteonCmd class."""
        super().__init__()
        self.stdin = stdin
        self.stdout = stdout
        self._prompt_base = "pyinsteon"
        if menu:
            self.prompt = f"{self._prompt_base} - {menu}: "
        else:
            self.prompt = f"{self._prompt_base}: "
        self._log_prefix = "STDOUT: "
        self._input_prefix = "INPUT"
        self._background_prefix = "BACKGROUND"
        self.loop = loop
        self.workdir = None

        self.device = None
        self.host = None
        self.username = None
        self.hub_version = 2
        self.port = 25105
        self.mock = False

        self._set_command_line_values(args)

        self._add_filter()

        if hasattr(args, "workdir"):
            self.workdir = args.workdir
            if hasattr(args, "logging") and self.workdir:
                self.do_log_to_file("y", self.workdir)

    def _set_command_line_values(self, args):
        """Set up the connection values from the command line arguments."""
        try:
            self.device = args.device if args.device else None
        except AttributeError:
            pass
        try:
            self.username = args.username if args.username else None
        except AttributeError:
            pass
        try:
            self.host = args.host if args.host else None
        except AttributeError:
            pass
        try:
            self.hub_version = args.hub_version if args.hub_version else 2
        except AttributeError:
            pass
        try:
            self.port = (
                args.port if args.port else DEFAULT_HUB_PORT.get(self.hub_version)
            )
        except AttributeError:
            pass

        if hasattr(args, "verbose") and args.verbose:
            self._setup_logging(logging.DEBUG)
        else:
            self._setup_logging(logging.INFO)

    async def async_cmdloop(self, *intro):
        """Replace standard cmdloop to make this async.

        Repeatedly issue a prompt, accept input, parse an initial prefix
        off the received input, and dispatch to action methods, passing them
        the remainder of the line as argument.

        Can be called with `async_cmdloop()` or `async_cmdloop("intro string")` or
        `async_cmdloop(*["intro line 1", "intro_line2", ...])`
        """

        self.preloop()
        if not self.stdin and not self.stdout:
            self.stdin, self.stdout = await stdio(loop=self.loop)
            if sys.platform != "win32":
                patch_stdin_stdout(self.stdin, self.stdout)

        try:
            for intro_str in intro:
                self.stdout.write(intro_str + "\n")
            stop = None
            while not stop:
                if self.cmdqueue:
                    line = self.cmdqueue.pop(0)
                else:
                    line = await self._input(prompt=self.prompt)
                    if not line:
                        line = ""
                line = self.precmd(line)
                try:
                    active_task = asyncio.create_task(self.async_onecmd(line))
                    self._set_active_task(active_task)
                    try:
                        stop = await active_task
                    except asyncio.CancelledError:
                        stop = None
                    self._set_active_task(None)
                # pylint: disable=broad-except
                except Exception as ex:
                    self._log_stdout("An error occured executing command.")
                    self._log_stdout(str(ex))
                    _LOGGING.debug(traceback.format_exc())
                stop = self.postcmd(stop, line)
            self.postloop()
        finally:
            pass

    def parseline(self, line):
        """Parse the full command line."""
        cmd, args, line = super().parseline(line)
        if args:
            args = args.split(" ")
            while "" in args:
                args.remove("")
        else:
            args = []
        return cmd, args, line

    async def async_onecmd(self, line):
        """Replace the onecmd method to make it async.

        Interpret the argument as though it had been typed in response
        to the prompt.
        """

        if not line:
            return self.emptyline()

        cmd, all_args, line = self.parseline(line)
        self._log_command(line)

        self.lastcmd = line

        try:
            func = getattr(self, "do_" + cmd)
        except AttributeError:
            try:
                func = getattr(self, "menu_" + cmd)
            except AttributeError:
                self._log_stdout(f"Invalid command: {cmd}")
                return
        args, kwargs, args_valid = self._parse_args(func, all_args)

        if not args_valid:
            return

        background = kwargs.get("background")
        if background:
            return asyncio.ensure_future(func(*args, **kwargs))
        if inspect.iscoroutinefunction(func) or inspect.isawaitable(func):
            return await func(*args, **kwargs)
        return func(*args, **kwargs)

    def emptyline(self):
        """Change default empty line to do nothing."""
        return

    @classmethod
    def start(cls):
        """Start the loop and the current command set."""
        parser = ArgumentParser(description=__doc__)
        parser.add_argument("--device", default="", help="Path to PLM device")
        parser.add_argument("--host", default="", help="Hostname or IP address of Hub")
        parser.add_argument("--username", default="", help="Username of Hub")
        parser.add_argument(
            "--hub_version", default="", help="Version of the Hub (1 or 2)"
        )
        parser.add_argument(
            "--port",
            default="",
            help="Hub IP port (Default 25105 for version 2 or 9761 for version 1)",
        )
        parser.add_argument(
            "-v", "--verbose", action="count", help="Set logging level to verbose"
        )
        parser.add_argument("-l", "--logging", action="count", help="Log to file")
        parser.add_argument(
            "--workdir",
            default="",
            help="Working directory for reading and saving device information.",
        )
        args = parser.parse_args()

        set_loop()
        loop = asyncio.get_event_loop()

        intro = [
            "",
            "The command line module for pyinsteon is designed to test devices and perform certain common functions.",
            "",
            "Enter `help` to see a list of commands.",
            "Most commands can be called with `--background` or `-b` as the first parameter to run the command asyncronously in the background."
            "For example:",
            "      status --backround aa.bb.cc",
            "",
        ]

        try:
            loop.add_signal_handler(signal.SIGINT, cls._cancel_active_task)
        except NotImplementedError:
            signal.signal(signal.SIGINT, cls._cancel_active_task)

        loop.run_until_complete(cls(loop, args).async_cmdloop(*intro))

    async def start_menu(self):
        """Start the menu."""
        await self.async_cmdloop()

    # pylint: disable=arguments-differ
    def do_help(self, arg=""):
        """List available commands with "help" or detailed help with "help cmd"."""
        menus = self._get_menus()
        if arg and arg in self._get_menus():
            try:
                func = getattr(self, "help_" + arg)
            except AttributeError:
                try:
                    doc = getattr(self, "menu_" + arg).__doc__
                    if doc:
                        self.stdout.write(f"{doc}\n")
                        return
                except AttributeError:
                    pass
                # pylint: disable=consider-using-f-string
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
            return
        if arg:
            self.stdout.write("\n")

        super().do_help(arg)
        if arg:
            try:
                func = getattr(self, "do_" + arg)
                full_args_spec = inspect.getfullargspec(func)
                if "background" in full_args_spec.args:
                    self.stdout.write(
                        "        --background | -b: Run in the background\n"
                    )
            except AttributeError:
                pass
            self.stdout.write("\n")
            return

        self.stdout.write("\n")
        menu_header = "List of available submenues:"
        self.print_topics(menu_header, menus, 15, 80)
        self.stdout.write("\n")

    def do_list_devices(self, log_stdout=None, background=False):
        """List all devices.

        Usage:
            list_devices [--background | -b]
        """
        log_stdout("Address   Cat   Subcat Description")
        log_stdout(
            "--------  ----- ------ ------------------------------------------------------------------------"
        )
        for addr in devices:
            device = devices[addr]
            log_stdout(
                f"{addr}  {device.cat!r}  0x{int(device.subcat):02x}   {device.description}"
            )
        log_stdout(f"Total devices: {len(devices)}")

    async def do_log_to_file(self, logging_mode, workdir=None):
        """Start logging to file.

        Usage:
            log_to_file logging_mode [location]

        logging_mode: y | n
        location: Location to store the log file (default is current directory)

        The log file name is `pyinsteon_tools.log`.
        """
        try:
            logging_mode = await self._ensure_bool(
                logging_mode, "Log mode", True, self._log_stdout
            )
        except ValueError:
            self._log_stdout("A valid value for log mode is required")
            return

        root_logger = logging.getLogger()
        if not logging_mode:
            for handler in root_logger.handlers:
                if handler.get_name() == FILE_LOG_HANDLER:
                    root_logger.removeHandler(handler)
            return

        if workdir:
            self.workdir = workdir

        if not self.workdir and not workdir:
            self.workdir = await self._get_workdir()

        log_file = os.path.join(self.workdir, LOG_FILE_NAME)
        file_formatter = logging.Formatter(
            "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
        )
        file_handler = logging.FileHandler(log_file, mode="a")
        file_handler.setFormatter(file_formatter)
        file_handler.set_name(FILE_LOG_HANDLER)
        file_handler.addFilter(StripPrefixFilter(self._input_prefix))
        file_handler.addFilter(StripPrefixFilter(self._background_prefix))
        root_logger.addHandler(file_handler)

    async def do_save_devices(self, workdir=None, log_stdout=None, background=False):
        """Save devices to the working directory.

        Usage:
            save_devices [--background | -b] workdir

        workdir: Directory where the saved device file is located
        """
        if workdir is None:
            workdir = self.workdir if self.workdir else "."

        self.workdir = workdir

        if not self.workdir:
            if background:
                log_stdout("A value for the working directory is required.")
                return
            self.workdir = await self._get_workdir()

        if not self.workdir:
            return

        await devices.async_save(workdir=self.workdir)

    async def do_back(self):
        """Go back to the previous menu.

        Usage:
            back
        """
        if self.prompt[:-2] != self._prompt_base:
            return -1

    async def do_exit(self):
        """Exit the tool.

        Usage:
            exit
        """
        await self.do_log_to_file("n")
        return True

    async def do_set_log_level(self, level):
        """Set the log level to INFO (i) or VERBOSE (v).

        Usage:
            set_log_level level

        level: Can be one of the following:
            i: Info
            v: Debug
            m: Show messages
            t: Show topics
        """

        options = ["i", "v", "m", "t", "n"]
        if level is not None:
            level = level.lower()

        level = await self._ensure_string(
            level,
            options,
            "Log level (i=info, v=verbose, m=messages, t=topics)",
            True,
            self._log_stdout,
        )

        root_logger = logging.getLogger()
        if level == "i":
            root_logger.setLevel(logging.INFO)
            message_logger = logging.getLogger("pyinsteon.messages")
            message_logger.setLevel(logging.ERROR)
            topic_logger = logging.getLogger("pyinsteon.topics")
            topic_logger.setLevel(logging.ERROR)
        elif level == "v":
            root_logger.setLevel(logging.DEBUG)
        elif level == "m":
            message_logger = logging.getLogger("pyinsteon.messages")
            message_logger.setLevel(logging.DEBUG)
        elif level == "t":
            topic_logger = logging.getLogger("pyinsteon.topics")
            topic_logger.setLevel(logging.DEBUG)

    async def do_debug_device(self, address, logging_mode=True):
        """Place a device into debug mode to log all topics for the device.

        usage:
            debug_device address [logging_mode]

        logging_mode: y | n  (default is y)

        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=True,
                log_stdout=self._log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                raise ValueError
            address = addresses[0]
        except ValueError:
            self._log_stdout("Invalid device address or device not found")
            return
        try:
            logging_mode = await self._ensure_bool(
                logging_mode, "Log mode", True, self._log_stdout
            )
        except ValueError:
            self._log_stdout("A valid value for log mode is required")
            return

        logger = logging.getLogger(f"pyinsteon.{address.id}")
        log_level = logging.DEBUG if logging_mode else logging.INFO
        logger.setLevel(log_level)

    async def do_status(self, address, reload, log_stdout=None, background=False):
        """Display the status of a device.

        Usage:
            status [--background | -b] <address>|all y|n

        address: Enter a single address or all for all devices
        reload: y to request the device to send status or
                n to use the current known status
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                if background:
                    raise ValueError
                return
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        try:
            refresh_current = await self._ensure_bool(
                reload, "Reload", not background, log_stdout
            )
            if refresh_current is None:
                if background:
                    raise ValueError
                return
        except ValueError:
            log_stdout("Invalid value for reload")
            return

        tasks = []

        for device_address in addresses:
            if devices[device_address] != devices.modem:
                if not refresh_current:
                    for group in devices[device_address].groups:
                        if devices[device_address].groups[group].value is None:
                            refresh_current = True
                            break
                if refresh_current:
                    tasks.append(devices[device_address].async_status())
        if tasks:
            await asyncio.gather(*tasks)

        for device_address in addresses:
            if devices[device_address] != devices.modem:
                self._print_device_status(log_stdout, device_address)

    async def do_monitor_mode(self):
        """Enter monitoring mode.

        Usage:
            monitor_mode
        """
        self._log_stdout("Press enter to exit monitor mode")
        self._remove_filter()
        await self.stdin.readline()
        self._add_filter()

    def _print_device_status(self, log_stdout, address):
        """Print device status to log."""
        device = devices[address]
        log_stdout("Address  Group State Name      Value")
        log_stdout("-------- ----- --------------- -----")
        for group_id in device.groups:
            group = device.groups[group_id]
            log_stdout(
                f"{device.address} {group_id:>3d}   {group.name:15.15s} {str(group.value):>5s}"
            )

    def _async_run(self, func, *args, **kwargs):
        """Run a function in the event loop."""
        self.loop.run_until_complete(func(*args, **kwargs))

    async def _get_connection_params(self, password):
        """Ensure connectoin parameters are filled."""
        if self.host:
            modem_type = "hub"
        elif self.device:
            modem_type = "plm"
        else:
            modem_type = None

        modem_type = await self._ensure_string(
            modem_type,
            ["plm", "hub"],
            "Modem type (plm or hub)",
            True,
            self._log_stdout,
        )
        if not modem_type:
            return None

        if modem_type == "plm":
            self.device = await self._ensure_string(
                self.device,
                None,
                "Device (i.e. /dev/ttyUSB0 or COM5)",
                True,
                self._log_stdout,
            )
            self._log_input(f"    Device: {self.device}")
            return

        self.host = await self._ensure_string(
            self.host, None, "Hub IP address or hostname", True, self._log_stdout
        )
        if not self.host:
            return

        self.username = await self._ensure_string(
            self.username, None, "Username", True, self._log_stdout
        )
        if not self.username:
            return

        self._log_command(f"Password in: {password}")
        if not password:
            self._log_command("Need password")
            if hasattr(self.stdin, "not_tty"):
                self._log_command("Using standard input method")
                # Used for test purposes to allow input via a stream vs a tty device
                password = await self._input("Password")
            else:
                password = getpass.getpass(prompt="Hub password: ")
            if not password:
                raise ValueError("Password is required")

        self.hub_version = await self._ensure_int(
            self.hub_version,
            [1, 2],
            "Hub version",
            True,
            self._log_stdout,
            default=2,
        )

        self.port = await self._ensure_int(
            self.port,
            None,
            "Hub port",
            True,
            self._log_stdout,
            default=DEFAULT_HUB_PORT[self.hub_version],
        )
        return password

    def postcmd(self, stop, line):
        """Perform post command execution."""
        if line == "exit":
            return True
        return False

    async def _input(self, prompt=""):
        """Asyncronous input of a line."""
        self.stdout.write(prompt)
        await self.stdout.drain()
        line = await self.stdin.readline()
        return line.strip("\r\n")

    def _setup_logging(self, level):
        """Set up the initial console logging."""

        root_logger = logging.getLogger()

        # remove any existing logging that are not our handlers
        # add filter to existing stdout handler
        found_our_handlers = False
        for handler in root_logger.handlers:
            if handler.get_name() not in [STDOUT_LOG_HANDLER, FILE_LOG_HANDLER]:
                logging.getLogger().removeHandler(handler)
            else:
                found_our_handlers = True

        if found_our_handlers:
            return

        stdout_formatter = logging.Formatter("%(message)s")
        console_handler = logging.StreamHandler(self.stdout)
        console_handler.setFormatter(stdout_formatter)
        console_handler.set_name(STDOUT_LOG_HANDLER)
        root_logger.addHandler(console_handler)
        root_logger.setLevel(level)

    async def _call_next_menu(self, menu, name=None):
        """Start the next menu."""
        cmd_args = CmdArgs(
            self.workdir,
            self.device,
            self.username,
            self.host,
            self.hub_version,
            self.port,
        )
        await menu(
            self.loop, cmd_args, name, stdin=self.stdin, stdout=self.stdout
        ).start_menu()

    def _add_filter(self):
        """Add a filter for the current menu."""
        root_logger = logging.getLogger()
        found_prompt = False
        found_prefix = False
        found_input = False
        found_background = False
        for handler in root_logger.handlers:
            if handler.get_name() == STDOUT_LOG_HANDLER:
                for my_filter in handler.filters:
                    if hasattr(my_filter, "prefix") and my_filter.prefix == self.prompt:
                        found_prompt = True
                    if (
                        hasattr(my_filter, "prefix")
                        and my_filter.prefix == self._log_prefix
                    ):
                        found_prefix = True
                    if (
                        hasattr(my_filter, "input_prefix")
                        and my_filter.prefix == self._input_prefix
                    ):
                        found_input = True
                    if (
                        hasattr(my_filter, "prefix")
                        and my_filter.prefix == self._background_prefix
                    ):
                        found_background = True
                if not found_prompt:
                    handler.addFilter(NoStdoutFilter(self.prompt))
                if not found_prefix:
                    handler.addFilter(StdoutFilter(self._log_prefix))
                if not found_input:
                    handler.addFilter(NoStdoutFilter(self._input_prefix))
                if not found_background:
                    handler.addFilter(NoStdoutFilter(self._background_prefix))

    def _remove_filter(self):
        """Remove the stdout filter to enable log output."""
        root_logger = logging.getLogger()
        found_prefix_filter = False
        for handler in root_logger.handlers:
            if handler.get_name() == STDOUT_LOG_HANDLER:
                for my_filter in handler.filters:
                    if (
                        hasattr(my_filter, "prefix")
                        and my_filter.prefix == self._log_prefix
                    ):
                        found_prefix_filter = True
                        stdout_filter = my_filter
                        break
            if found_prefix_filter:
                handler.removeFilter(stdout_filter)

    def _log_command(self, line):
        """Log the command to the log file if the log file is active."""
        # This works because there is a filter on stdout for
        # any string begining with self.prompt
        output = f"{self.prompt}{line}"
        _LOGGING.info(output)

    def _log_background(self, line):
        """Log the command to the log file if the log file is active."""
        # This works because there is a filter on stdout for
        # any string begining with self.prompt
        output = f"{self._background_prefix}{line}"
        _LOGGING.info(output)

    def _log_stdout(self, line):
        """Log a message to standard out."""
        output = f"{self._log_prefix}{line}"
        _LOGGING.info(output)

    def _log_input(self, line):
        """Log a message to standard out."""
        output = f"{self._input_prefix}{line}"
        _LOGGING.info(output)

    async def _get_int(self, prompt, default=None, values=None):
        """Get an integer value."""
        if default is not None:
            prompt = f"{prompt} (Default {default}): "
        else:
            prompt = f"{prompt}: "
        while True:
            value = await self._input(prompt)
            self._log_input(f"    {prompt}: {value}")
            if value == "":
                return default
            try:
                value = int(value)
                if values and value not in values:
                    raise ValueError
                return value
            except ValueError:
                response = "Must be a number."
                if values:
                    response = f"{response} Acceptable values {values}."
                self._log_stdout(response)

    async def _get_float(self, prompt, default=None, maximum=None, minimum=None):
        """Get a floating point value."""
        if default is not None:
            prompt = f"{prompt} (Default {default}): "
        else:
            prompt = f"{prompt}: "
        while True:
            value = await self._input(prompt)
            self._log_input(f"    {prompt}: {value}")
            if value == "":
                return default
            try:
                value = float(value)
                if minimum is not None and value < minimum:
                    raise ValueError
                if maximum is not None and value > maximum:
                    raise ValueError
                return value
            except ValueError:
                response = "Must be a number."
                if maximum is not None and minimum is not None:
                    response = f"{response} (Max: {maximum}  Min: {minimum})."
                elif maximum is not None:
                    response = f"{response} (Max: {maximum})."
                elif minimum is not None:
                    response = f"{response} (Min: {minimum})."
                self._log_stdout(response)

    async def _get_char(
        self, prompt, default=None, values=None, ignore_case=True, hex_string=False
    ):
        """Get a character string value."""
        if default is not None:
            prompt = f"{prompt} (Default {default.upper()}): "
        else:
            prompt = f"{prompt}: "
        if ignore_case and values:
            values = [val.lower() for val in values]

        while True:
            value = await self._input(prompt)
            self._log_input(f"    {prompt}: {value}")
            if hex_string and value.startswith("0x"):
                value = value[2:]

            if value == "":
                return default

            if not values:
                return value

            if ignore_case and value.lower() in values:
                return value.lower()

            if value in values:
                return value

            self._log_stdout(f"Acceptable values {values}.")

    async def _get_workdir(self):
        """Input the value for the workdir."""
        self._log_stdout("The working directory stores the list of identified devices.")
        self._log_stdout(
            "Enter a working directory where the saved file is (and will be saved to after loading.)"
        )
        while True:
            workdir = await self._input(
                "Working directory (enter . for current director): "
            )
            self._log_input(f"    Working Directory: {workdir}")

            if workdir == "":
                return None

            if workdir == ".":
                return os.getcwd()

            if os.path.exists(workdir):
                return workdir

            self._log_stdout("Directory does not exist")

    async def _get_address(
        self,
        allow_all=True,
        match_device=True,
        prompt="Address",
    ):
        """Get the address of a device or all devices."""
        if allow_all:
            input_prompt = f"{prompt} or all for all device addresses: "
        else:
            input_prompt = prompt

        while True:
            address = await self._input(input_prompt)
            self._log_input(f"    {input_prompt}: {address}")

            if address == "":
                return []

            try:
                addresses = validate_address(
                    address, allow_all, match_device, self._log_stdout
                )
                return addresses
            except ValueError:
                address = None

    def _print_aldb_out(self, addresses, unused, log_stdout):
        """Print the ALDB to the log."""
        for address in addresses:
            device = devices[address]
            records = [
                device.aldb[mem_addr]
                for mem_addr in device.aldb
                if (unused or device.aldb[mem_addr].is_in_use)
            ]
            self._print_aldb_output(log_stdout, device, records)

    def _print_aldb_output(self, log_stdout, device="", records=1, **kwargs):
        log_stdout("")
        log_stdout(f"Device: {device.address}  Load Status: {str(device.aldb.status)}")
        log_stdout("RecID In Use Mode HWM Group Address  Data 1 Data 2 Data 3")
        log_stdout("----- ------ ---- --- ----- -------- ------ ------ ------")
        for rec in records:
            in_use = "Y" if rec.is_in_use else "N"
            link_mode = "C" if rec.is_controller else "R"
            hwm = "Y" if rec.is_high_water_mark else "N"
            line = f" {rec.mem_addr:04x}    {in_use:s}     {link_mode:s}   {hwm:s}    {rec.group:3d} {rec.target}   {rec.data1:3d}   {rec.data2:3d}   {rec.data3:3d}"
            log_stdout(line)
        log_stdout("")

    @classmethod
    def _set_active_task(cls, task):
        """Set the active task."""
        cls._active_task = task

    @classmethod
    def _cancel_active_task(cls, *args, **kwargs):
        """Cancel the active task."""
        if cls._active_task is not None:
            cls._active_task.cancel()

    async def _ensure_byte(self, value, name, ask_value, log_stdout, default=None):
        """Ensure a given `value` is a single bite between 0 - 255.

        Returns `None` or default if value is `None` and `ask_value` is `False`.
        Raises `ValueError` if `value` is not a valid byte and `ask_value` is `False`.
        """
        if value is None and not ask_value:
            return default

        try:
            value = int(value) if value is not None else None
            if value not in range(0, 256):
                raise ValueError
        except ValueError as ex:
            log_stdout(f"Invalid {name} value")
            if not ask_value:
                raise ValueError from ex
            value = None
        if value is None:
            value = await self._get_int(name, default=default, values=range(0, 256))
        return value

    async def _ensure_bool(
        self,
        value,
        name,
        ask_value,
        log_stdout,
        true_val=None,
        values=None,
        default=None,
    ):
        """Compare a given `value` against `true_val`.

        Returns `None` if value is `None` and `ask_value` is `False`.
        Raises `ValueError` if `value` is not in the list of acceptable values
        and `ask_value` is `False`.
        """

        if value is None and not ask_value:
            return default

        if true_val is None:
            true_val = "y"

        if values is None:
            values = ["y", "n"]

        if value is not None:
            if str(value).lower() in values:
                return str(value).lower() == true_val
            log_stdout(f"Invalid {name} value")
            if not ask_value:
                raise ValueError
            value = None

        default_char = true_val if default else ""
        val = await self._get_char(name, values=values, default=default_char)
        if val == "":
            return default
        return val == true_val

    async def _ensure_int(
        self, value, values, name, ask_value, log_stdout, default=None
    ):
        """Ensure a value is an int within the acceptable values`.

        Returns `default` if value is `None` and `ask_value` is `False`.
        Raises `ValueError` if `value` is not in the list of acceptable values
        and `ask_value` is `False`.
        """
        if value is None and not ask_value:
            return default

        if value is not None:
            try:
                value = int(value)
                if not values or value in values:
                    return value
                raise ValueError
            except (ValueError, TypeError):
                log_stdout(f"Invalid {name} value")
                value = None

        if not ask_value:
            raise ValueError

        return await self._get_int(name, values=values, default=default)

    async def _ensure_float(
        self, value, minimum, maximum, name, ask_value, log_stdout, default=None
    ):
        """Ensure a value is a floting point number within the acceptable values`.

        Returns `None` if value is `None` and `ask_value` is `False`.
        Raises `ValueError` if `value` is not in the list of acceptable values
        and `ask_value` is `False`.
        """
        if value is None and not ask_value:
            return default

        if value is not None:
            try:
                value = float(value)
                if minimum is not None and value < minimum:
                    raise ValueError
                if maximum is not None and value > maximum:
                    raise ValueError
                return value
            except (ValueError, TypeError):
                log_stdout(f"Invalid {name} value")
                value = None

        if not ask_value:
            raise ValueError

        return await self._get_float(
            name, default=default, minimum=minimum, maximum=maximum
        )

    async def _ensure_string(
        self, value, values, name, ask_value, log_stdout, default=None
    ):
        """Ensure a value is a string within the acceptable values`.

        Returns `default` if `default` is set and value is `None` and `ask_value` is `False`.

        Raises `ValueError` if
        - `value` is not in the list of acceptable values and `ask_value` is `False`.
        - `value` is `None` and `default` is not set and `ask_value` is `False`
        """
        if value is None and not ask_value:
            return default

        if value is not None:
            value = str(value)
            if not values or value in values:
                return value
            log_stdout(f"Invalid {name} value")
            value = None

        if not ask_value:
            raise ValueError

        return await self._get_char(name, values=values, default=default)

    async def _ensure_address(
        self,
        address,
        name,
        ask_value,
        log_stdout,
        allow_all=True,
        match_device=True,
    ) -> List[Address]:
        """Ensure a value is a proper device address.

        Returns `[]` if the value is `None` and `ask_value` is `False`

        Returns a list of addresses if value is `all` and allow_all is `True`

        Raises `ValueError` if
        - `value` is not a valid address and `ask_value` is `False`
        - `value` does not match a known device and `match_device` is `True` and `ask_value` is `False`
        - `value` is `all` and `allow_all` is `False` and `ask_value` is `False`
        """
        if address is None and not ask_value:
            return []

        try:
            return validate_address(address, allow_all, match_device, log_stdout)
        except ValueError:
            if not ask_value:
                raise

        return await self._get_address(
            allow_all=allow_all,
            match_device=match_device,
            prompt=name,
        )

    def _get_menus(self):
        """Return the list of menu commands."""
        menus = []
        for method in self.get_names():
            if method.startswith("menu_"):
                menus.append(method[len("menu_") :])
        return menus

    def _parse_args(self, func, all_args):
        def to_args_kwargs(arg_in, f_args, args, kwargs, found_kwarg):
            loc_eq = arg_in.find("=")
            if loc_eq != -1:
                found_kwarg = True
                key = arg_in[:loc_eq]
                val = arg_in[loc_eq + 1 :]
                kwargs[key] = val

            # We found a kwarg previously and now we don't have a kwarg
            # That is an issue!
            elif found_kwarg:
                raise ValueError
            else:
                args.append(arg_in)
                try:
                    f_args.pop(0)
                except IndexError:
                    pass
            return f_args, args, kwargs, found_kwarg

        full_args_spec = inspect.getfullargspec(func)
        f_args = full_args_spec.args
        f_kwargs = full_args_spec.kwonlyargs
        has_vargs = full_args_spec.varargs is not None
        has_kwargs = full_args_spec.varkw is not None
        len_defaults = len(full_args_spec.defaults or [])
        if "self" in f_args:
            f_args.remove("self")

        found_kwarg = False
        args = []
        kwargs = {}
        background = False
        for item in ("--background", "-b"):
            try:
                all_args.remove(item)
                background = True
            except ValueError:
                pass
        # Set up background and log_stdout kwargs if possible
        if "background" in f_args or "background" in f_kwargs:
            kwargs["background"] = background
            if len_defaults >= 2:
                len_defaults -= 2
            if background:
                kwargs["log_stdout"] = self._log_background
            else:
                kwargs["log_stdout"] = self._log_stdout
            if "background" in f_args:
                f_args.remove("background")
            if "log_stdout" in f_args:
                f_args.remove("log_stdout")

        elif background:
            self._log_stdout("Command cannot be run in the background.")
            background = False

        len_f_args = len(f_args)
        for _ in range(0, len_f_args):
            try:
                arg_in = all_args[0]
                f_args, args, kwargs, found_kwarg = to_args_kwargs(
                    arg_in, f_args, args, kwargs, found_kwarg
                )
                all_args.pop(0)
            except IndexError:
                break
            except ValueError:
                self._log_stdout("Keyword argument before non-keyword argument")
                return args, kwargs, False

        vargs = []
        for arg_in in all_args:
            try:
                f_args, vargs, kwargs, found_kwarg = to_args_kwargs(
                    arg_in, f_args, vargs, kwargs, found_kwarg
                )
            except ValueError:
                self._log_stdout("Keyword argument before non-keyword argument")
                return args, kwargs, False

        if vargs and not has_vargs:
            self._log_stdout("Too many arguements")
            return args, kwargs, False

        args += vargs

        # Test if there are items in kwargs that cannot be handled
        for arg in kwargs:
            if arg in ["background", "log_stdout"]:
                continue
            if arg not in f_args and not has_kwargs:
                self._log_stdout(f"Invalid keyword argument {arg}")
                return args, kwargs, False
            if arg in f_args:
                f_args.remove(arg)

        # Test if we have enough args to continue
        if len(f_args) > len_defaults and background:
            self._log_stdout("Missing arguments required to run in background")
            return args, kwargs, False

        # Pad missing args with `None`.
        for _ in range(0, max(0, len(f_args) - len_defaults)):
            args.append(None)

        return args, kwargs, True

    async def _ensure_arg_value(
        self,
        arg,
        value,
        ask_value,
        log_stdout,
        values=None,
        default=None,
        true_val=None,
        require_two=True,
    ):
        """Convert an argument value from string to argument type.

        Ensures a value for an argument is valid. Uses the ARG_TYPE list to determine the argument type.

        If the argument is `buttons`, the `requires_two` argument determines if two or more buttons is required.
        """

        arg_type = ARG_TYPES.get(arg)
        # pylint: disable=too-many-nested-blocks
        if arg_type is not None:
            try:
                if arg_type == bool:
                    return await self._ensure_bool(
                        value,
                        arg,
                        ask_value,
                        log_stdout,
                        default=default,
                        values=values,
                        true_val=true_val,
                    )

                if arg_type == bytes:
                    return await self._ensure_byte(
                        value, arg, ask_value, log_stdout, default=default
                    )

                if isinstance(arg_type, range):
                    values = values if values is not None else arg_type
                    return await self._ensure_int(
                        value=value,
                        values=values,
                        name=arg,
                        ask_value=ask_value,
                        log_stdout=log_stdout,
                        default=default,
                    )

                if arg.lower() == "buttons":
                    if value is None and not ask_value:
                        return default

                    buttons = []
                    values = values if values else arg_type[0]
                    for button in range(0, len(values)):
                        input_button = None
                        if value:
                            try:
                                input_button = value[button]
                            except IndexError:
                                pass
                        try:
                            button = await self._ensure_int(
                                value=input_button,
                                values=values,
                                name="Button",
                                ask_value=ask_value,
                                log_stdout=log_stdout,
                            )
                        except ValueError as ex:
                            log_stdout("Invalid button number")
                            if not ask_value:
                                raise ValueError from ex
                        if button is None and len(buttons) < 2 and require_two:
                            log_stdout("At least two buttons are required")
                            if not ask_value:
                                raise ValueError
                        if button is None:
                            break
                        buttons.append(button)
                    return buttons

                if arg_type == Address:
                    return await self._ensure_address(
                        address=value,
                        name=arg,
                        ask_value=ask_value,
                        log_stdout=log_stdout,
                        allow_all=False,
                        match_device=False,
                    )

                # enum type:
                log_stdout("\nValid values are:")
                options = []
                for option in arg_type:
                    log_stdout(f"    {option}: {int(option)}")
                    options.append(option)
                return await self._ensure_int(
                    value, options, arg, ask_value, log_stdout, default=default
                )

            except (TypeError, ValueError) as ex:
                log_stdout(f"Invalid value for {arg}")
                raise ValueError from ex
        return value

    async def _ensure_hex_byte(self, value, name, ask_value, log_stdout, values=None):
        """Ensure a value is a hex byte string.

        Returns a bytes value equivalent of the hex string.

        byte_len is the number of bytes. Default is 1 byte (i.e. 0x01, 0xff).
        """

        def hex_str_to_int(value):
            """Convert a hex string into an integer."""
            if value and value.startswith("0x"):
                value = value[2:]
            byte_val = unhexlify(value)
            return int.from_bytes(byte_val, byteorder="big")

        convert_length = {
            range(0, 256): 2,
            range(256, 65536): 4,
            range(65536, 16777216): 6,
        }
        if value is None and not ask_value:
            raise ValueError

        values_text = []

        # pylint: disable=too-many-nested-blocks
        while True:
            if value is not None:
                try:
                    int_value = hex_str_to_int(value)
                    if not values or int_value in values:
                        return int_value
                    value = None
                except (BinasciiError, TypeError, ValueError, IndexError):
                    log_stdout(f"Invalid value for {name}")
                    value = None
            if ask_value:
                if values and not values_text:
                    val_length = 6
                    for val in values:
                        for val_range, convert_len in convert_length.items():
                            if val in val_range:
                                val_length = convert_len
                                break
                        if val_length == 2:
                            values_text.append(f"{val:02x}")
                        elif val_length == 4:
                            values_text.append(f"{val:04x}")
                        else:
                            values_text.append(f"{val:06x}")
                value = await self._get_char(name, values=values_text, hex_string=True)
                if value is None:
                    return None
            else:
                raise ValueError



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...topics import EXTENDED_READ_WRITE_ALDB
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class ReadALDBCommandHandler(DirectCommandHandlerBase):
    """Handle sending a read request for ALDB records."""

    def __init__(self, address: Address):
        """Init the ReadALDBCommandHandler."""
        super().__init__(topic=EXTENDED_READ_WRITE_ALDB, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, mem_addr: int = 0x0000, num_recs: int = 0):
        """Send ALDB read message asyncronously."""
        return await super().async_send(
            action=0x00, mem_addr=mem_addr, num_recs=num_recs
        )

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(response=0)

    def _update_subscribers_on_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers on DIIRECT NAK received."""
        self._call_subscribers(response=cmd2)

    def _update_subscribers_on_direct_nak(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers on DIIRECT NAK received."""
        self._call_subscribers(response=cmd2)



"""Manage inbound ON command from device."""

from ...address import Address
from ...constants import RampRate
from ...data_types.user_data import UserData
from ...topics import OFF_AT_RAMP_RATE_INBOUND
from .broadcast_command import BroadcastCommandHandlerBase


class OffAtRampRateInbound(BroadcastCommandHandlerBase):
    """Off At Ramp Rate command inbound."""

    def __init__(self, address: Address, group: int):
        """Init the OnAtRampRateInbound class."""
        super().__init__(topic=OFF_AT_RAMP_RATE_INBOUND, address=address, group=group)

    def _handle_message_received(
        self, cmd1: int, cmd2: int, target: Address, user_data: UserData, hops_left: int
    ):
        """Handle the OFF command from a device."""
        on_level = ((cmd2 >> 4) * 16) + 15
        ramp_rate_byte = ((cmd2 & 0x0F) * 2) + 1
        ramp_rate = RampRate(ramp_rate_byte)
        self._call_subscribers(on_level=on_level, ramp_rate=ramp_rate)



"""Wet / Dry state."""

from ..address import Address
from .group_base import GroupBase


class Dry(GroupBase):
    """Dry state."""

    def __init__(
        self,
        name: str,
        address: Address,
        group: int = 0,
        default: bool = None,
        status_type: int = 0,
    ):
        """Init the WetDry class."""
        super().__init__(
            name, address, group, default, value_type=bool, status_type=status_type
        )

    # pylint: disable=arguments-differ
    def set_value(self, dry):
        """Set the value of the state from the handlers."""
        self.value = bool(dry)


class Wet(GroupBase):
    """Wet state."""

    def __init__(
        self, name: str, address: Address, group: int = 0, default: bool = None
    ):
        """Init the WetDry class."""
        super().__init__(name, address, group, default, value_type=bool)

    # pylint: disable=arguments-differ
    def set_value(self, dry):
        """Set the value of the state from the handlers."""
        self.value = not bool(dry)



"""Handle an outbound All-Link Broadcast message to a group."""

from abc import ABCMeta

from .. import (
    ack_handler,
    all_link_cleanup_ack_handler,
    all_link_cleanup_nak_handler,
    nak_handler,
)
from ...constants import MessageFlagType
from ..outbound_base import OutboundHandlerBase


class AllLinkCleanupCommandHandlerBase(OutboundHandlerBase):
    """Abstract base class for outbound All-Link clean-up message handling."""

    __meta__ = ABCMeta

    def __init__(self, topic, address, group):
        """Init the DirectCommandHandlerBase class."""
        super().__init__(
            topic=topic,
            address=address,
            group=group,
            message_type=MessageFlagType.ALL_LINK_CLEANUP,
        )

    async def async_send(self, **kwargs):
        """Send the command and wait for a direct_nak."""
        return await super().async_send(address=self._address, **kwargs)

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""
        await self._async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)

    @nak_handler
    async def async_handle_nak(self, cmd1, cmd2, user_data):
        """Handle the message NAK."""
        await self._async_handle_nak()

    @all_link_cleanup_ack_handler
    def handle_all_link_ack(self, target, cmd1, cmd2, user_data, hops_left):
        """Handle the All-LInk Command ACK."""

    @all_link_cleanup_nak_handler
    def handle_all_link_nak(self, target, cmd1, cmd2, user_data, hops_left):
        """Handle the All-LInk Command NAK."""



"""Manage Insteon Scenes."""
import asyncio
import json
import logging
from os import path
from typing import Dict, Union

import aiofiles
import voluptuous as vol

from .. import devices
from ..address import Address
from ..constants import ResponseStatus
from ..handlers.send_all_link_off import SendAllLinkOffCommandHandler
from ..handlers.send_all_link_on import SendAllLinkOnCommandHandler
from ..utils import multiple_status
from .device_link_manager import LinkInfo

SCENE_FILE = "insteon_scenes.json"
_LOGGER = logging.getLogger(__name__)
ControllerAddress = Address
ResponderAddress = Address
Group = int
_scene_names = {}

DeviceLinkSchema = vol.Schema(
    [
        {
            vol.Required("address"): str,
            vol.Required("data1"): int,
            vol.Required("data2"): int,
            vol.Required("data3"): int,
        }
    ]
)


def _json_key_to_int(key):
    """Ensure the Json key is an int."""
    if isinstance(key, dict):
        return {int(k): v for k, v in key.items()}
    return key


async def _get_scene_device_status(group: int):
    """Get the status of the devices in a scene."""
    scene = await async_get_scene(group)
    for addr in scene["devices"]:
        device = devices[addr]
        if device:
            await device.async_status()


async def async_trigger_scene_on(group):
    """Trigger an Insteon scene ON."""
    await SendAllLinkOnCommandHandler().async_send(group=group)
    await asyncio.sleep(2)
    await _get_scene_device_status(group)


async def async_trigger_scene_off(group):
    """Trigger an Insteon scene OFF."""
    await SendAllLinkOffCommandHandler().async_send(group=group)
    await asyncio.sleep(2)
    await _get_scene_device_status(group)


async def async_get_scenes(work_dir=None):
    """Return a list of scenes."""
    scenes: Dict[Group, Dict[str, Union[Dict[ResponderAddress, LinkInfo], str]]] = {}
    if work_dir:
        await async_load_scene_names(work_dir=work_dir)
    for addr in devices:
        device = devices[addr]
        if device == devices.modem:
            continue
        for rec in device.aldb.find(
            target=devices.modem.address, is_controller=False, in_use=True
        ):
            if rec.group == 0:
                continue
            if not scenes.get(rec.group):
                scenes[rec.group] = {}
                scenes[rec.group]["name"] = _scene_names.get(
                    rec.group, f"Insteon Scene {rec.group}"
                )
                scenes[rec.group]["group"] = rec.group
                scenes[rec.group]["devices"] = {}
            scene = scenes.get(rec.group)
            if not scene["devices"].get(device.address):
                scene["devices"][device.address] = []
            has_controller = False
            for _ in devices.modem.aldb.find(
                target=device.address, group=rec.group, is_controller=True, in_use=True
            ):
                has_controller = True
                break
            scene["devices"][device.address].append(
                LinkInfo(rec.data1, rec.data2, rec.data3, has_controller, True)
            )
    return scenes


async def async_get_scene(scene_num: int, work_dir: str = None):
    """Return a scenes."""
    scene: Dict[str, Union[Dict[ResponderAddress, LinkInfo], str]] = {}
    if work_dir:
        await async_load_scene_names(work_dir=work_dir)
    scene["name"] = _scene_names.get(scene_num, f"Insteon Scene {scene_num}")
    scene["group"] = scene_num
    scene["devices"] = {}
    for addr in devices:
        device = devices[addr]
        if device == devices.modem:
            continue
        for rec in device.aldb.find(
            target=devices.modem.address,
            is_controller=False,
            in_use=True,
            group=scene_num,
        ):
            if rec.group == 0:
                continue
            if not scene["devices"].get(device.address):
                scene["devices"][device.address] = []
            has_controller = False
            for _ in devices.modem.aldb.find(
                target=device.address, group=rec.group, is_controller=True, in_use=True
            ):
                has_controller = True
                break
            scene["devices"][device.address].append(
                LinkInfo(rec.data1, rec.data2, rec.data3, has_controller, True)
            )
    return scene


def set_scene_name(scene_num: int, name: str):
    """Set the friendly name of a scene."""
    _scene_names[scene_num] = name


async def async_load_scene_names(work_dir: str):
    """Load the scene names from the saved location."""
    json_file = {}
    scene_file = path.join(work_dir, SCENE_FILE)
    try:
        async with aiofiles.open(scene_file, "r") as afp:
            json_file = await afp.read()
        try:
            saved_devices = json.loads(json_file, object_hook=_json_key_to_int)
            _scene_names.clear()
            for scene_num, name in saved_devices.items():
                _scene_names[scene_num] = name
        except json.decoder.JSONDecodeError:
            _LOGGER.debug("Loading saved device file failed")
    except FileNotFoundError:
        _LOGGER.debug("Saved device file not found")


async def async_save_scene_names(work_dir: str):
    """Write the scenes to a file."""
    out_json = json.dumps(_scene_names, indent=2)
    scene_file = path.join(work_dir, SCENE_FILE)
    async with aiofiles.open(scene_file, "w") as afp:
        await afp.write(out_json)
        await afp.flush()


async def async_add_device_to_scene(
    address: Address,
    scene_num: int,
    data1: int = 255,
    data2: int = 0,
    data3: int = 1,
    delay_write: bool = False,
):
    """Add a device to a scene."""
    device = devices[address]
    modem = devices.modem
    device.aldb.add(
        group=scene_num,
        target=modem.address,
        controller=False,
        data1=data1,
        data2=data2,
        data3=data3,
    )
    modem.aldb.add(
        group=scene_num,
        target=device.address,
        controller=True,
        data1=int(device.cat),
        data2=device.subcat,
        data3=device.firmware if device.firmware is not None else 0,
    )
    if not delay_write:
        await device.aldb.async_write()
        await modem.aldb.async_write()


async def async_remove_device_from_scene(
    address: Address,
    scene_num: int,
    delay_write: bool = False,
):
    """Remove a device from a scene."""

    device = devices[address]
    modem = devices.modem
    for rec in device.aldb.find(
        group=scene_num, target=modem.address, is_controller=False, in_use=True
    ):
        device.aldb.remove(rec.mem_addr)
    for rec in modem.aldb.find(
        group=scene_num, target=device.address, is_controller=True, in_use=True
    ):
        modem.aldb.remove(rec.mem_addr)

    if not delay_write:
        await device.aldb.async_write()
        await modem.aldb.async_write()


async def async_add_or_update_scene(
    scene_num: int,
    links: DeviceLinkSchema,
    name: str = None,
    work_dir: str = None,
) -> int:
    """Create or update a scene with a list of devices and device link data."""
    try:
        DeviceLinkSchema(links)
    except vol.Error as exc:
        _LOGGER.error("Invalid DeviceLinkSchema format for links parameter.")
        raise exc

    updated_devices = []
    curr_scene = await async_get_scene(scene_num, work_dir=work_dir)

    if scene_num <= 0:
        scene_num = await _find_next_scene()

    # Remove all records from the current scene
    # Not a big deal since when we add them back they will not need to be writen again if they did not change
    updated_devices = _remove_scene_links(scene_num, curr_scene)

    # Add the devices from device_info param
    for link in links:
        device = devices[link["address"]]
        device.aldb.add(
            group=scene_num,
            target=devices.modem.address,
            controller=False,
            data1=link["data1"],
            data2=link["data2"],
            data3=link["data3"],
        )
        if device not in updated_devices:
            updated_devices.append(device)
        devices.modem.aldb.add(
            group=scene_num,
            target=device.address,
            controller=True,
            data1=int(device.cat),
            data2=device.subcat,
            data3=device.firmware if device.firmware is not None else 0,
        )

    result = await _async_write_scene_link_changes(updated_devices)
    if name:
        set_scene_name(scene_num=scene_num, name=name)
        if work_dir:
            await async_save_scene_names(work_dir)
    return scene_num, result


async def async_delete_scene(scene_num: int, work_dir: Union[str, None]):
    """Delete a scene."""
    curr_scene = await async_get_scene(scene_num)
    updated_devices = _remove_scene_links(scene_num, curr_scene)
    result = await _async_write_scene_link_changes(updated_devices)
    if result == ResponseStatus.SUCCESS:
        try:
            _scene_names.pop(scene_num)
        except KeyError:
            pass
        if work_dir:
            await async_save_scene_names(work_dir=work_dir)
    return result


def _remove_scene_links(scene_num: int, scene_data):
    """Mark all current scene links as unused.

    Does not write to the database.
    """
    updated_devices = []
    if scene_data:
        curr_device_info = scene_data["devices"]
        for addr, info_list in curr_device_info.items():
            device = devices[addr]
            for info in info_list:
                for rec in device.aldb.find(
                    is_controller=False,
                    group=scene_num,
                    target=devices.modem.address,
                    data3=info.data3,
                    in_use=True,
                ):
                    device.aldb.remove(rec.mem_addr)
                    if device not in updated_devices:
                        updated_devices.append(device)
                for modem_rec in devices.modem.aldb.find(
                    is_controller=True,
                    target=device.address,
                    group=scene_num,
                    in_use=True,
                ):
                    devices.modem.aldb.remove(modem_rec.mem_addr)
    return updated_devices


async def _async_write_scene_link_changes(device_list):
    """Write the ALDB for the devices changed."""
    results = []
    for device in device_list:
        _, failed = await device.aldb.async_write()
        result = ResponseStatus.FAILURE if failed else ResponseStatus.SUCCESS
        results.append(result)
    _, failed = await devices.modem.aldb.async_write()
    result = ResponseStatus.FAILURE if failed else ResponseStatus.SUCCESS
    results.append(result)
    return multiple_status(*results)


async def _find_next_scene():
    """Return the next available scene number."""
    next_scene = 20
    scenes = await async_get_scenes()
    while next_scene in scenes:
        next_scene += 1
    return next_scene



"""Derived property to hold an operating flag in an device property."""
from ..utils import bit_is_set, set_bit
from . import get_usable_value
from .derived_property import DerivedProperty


class OpFlagProperty(DerivedProperty):
    """Derived property to hold an operating flag in an device property."""

    def __init__(self, address, name, prop, prop_bit):
        """Init the RampRateProperty class."""
        super().__init__(address, name, bool, False, False)

        self._prop = prop
        self._prop_bit = prop_bit

    @property
    def value(self):
        """Return the ramp rate property value in seconds."""
        if self._prop.value is None:
            return None
        return bit_is_set(self._prop.value, self._prop_bit)

    @property
    def new_value(self):
        """Return the ramp rate property in seconds."""
        if self._prop.new_value is None:
            return None
        new_val = bit_is_set(self._prop.new_value, self._prop_bit)
        if new_val == self.value:
            return None
        return new_val

    @new_value.setter
    def new_value(self, value):
        """Set the ramp rate property in seconds."""
        if self._prop.value is None:
            new_value = 0
        else:
            new_value = get_usable_value(self._prop)

        if value is None:
            new_value = set_bit(new_value, self._prop_bit, self.value)
        else:
            new_value = set_bit(new_value, self._prop_bit, value)

        self._prop.new_value = new_value

    @property
    def is_loaded(self):
        """Return if the Operating flag has been loaded."""
        return self._prop.is_loaded

    @property
    def is_dirty(self):
        """Return if the Operating flag has been changed."""
        return self.new_value is not None



"""Message definitions."""

from collections import namedtuple

from ...address import Address
from ...constants import (
    AckNak,
    AllLinkMode,
    DeviceCategory,
    ImButtonEvents,
    ManageAllLinkRecordAction,
    MessageId,
    X10CommandType,
)
from ...data_types.all_link_record_flags import AllLinkRecordFlags
from ...data_types.im_config_flags import IMConfigurationFlags
from ...data_types.message_flags import MessageFlags
from ...data_types.user_data import UserData
from .message_definition import MessageDefinition

MessageField = namedtuple("MessageField", "name length type")


# MessageDefinition = namedtuple("MessageDefinition", "type fields")

# INSTEON Standard Message Received 0x50
FLD_STD_REC = [
    MessageField("address", 3, Address),
    MessageField("target", 3, Address),
    MessageField("flags", 1, MessageFlags),
    MessageField("cmd1", 1, int),
    MessageField("cmd2", 1, int),
]

# INSTEON Extended Message Received 0x51
FLD_EXT_REC = FLD_STD_REC.copy()
FLD_EXT_REC.append(MessageField("user_data", 14, UserData))

# X10 Send / Received 0x63 / 0x52
FLD_X10_SEND_REC = [
    MessageField("raw_x10", 1, int),
    MessageField("x10_flag", 1, X10CommandType),
]
FLD_X10_SEND_REC_ACK = FLD_X10_SEND_REC.copy()
FLD_X10_SEND_REC_ACK.append(MessageField("ack", 1, AckNak))

# ALL-Linking Completed 0x53
FLD_ALL_LINK_COMPLETE = [
    MessageField("link_mode", 1, AllLinkMode),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
    MessageField("cat", 1, DeviceCategory),
    MessageField("subcat", 1, int),
    MessageField("firmware", 1, int),
]

# IM Button Event Report 0x54
FLD_IM_BTN_EVENT_REPORT_REC = [MessageField("event", 1, ImButtonEvents)]

# IM User Reset Detected 0x55
FLD_USER_RESET_REC = []

# ALL-Link Cleanup Failure Report 0x56
FLD_ALL_LINK_CLEANUP_FAILURE = [
    MessageField("error", 1, int),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
]

# ALL-Link Record Response 0x57
FLD_ALL_LINK_RECORD_RESP = [
    MessageField("flags", 1, AllLinkRecordFlags),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
    MessageField("data1", 1, int),
    MessageField("data2", 1, int),
    MessageField("data3", 1, int),
]

# All-Link Cleanup Status Report 0x58
FLD_ALL_LINK_CLEANUP_REPORT_ACK = [MessageField("ack", 1, AckNak)]

# Read from EEPROM Response 0x59
FLD_READ_EEPROM_RESPONSE = [
    MessageField("mem_hi", 1, int),
    MessageField("mem_low", 1, int),
    MessageField("flags", 1, AllLinkRecordFlags),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
    MessageField("data1", 1, int),
    MessageField("data2", 1, int),
    MessageField("data3", 1, int),
]

# Get IM Info 0x60
FLD_GET_IM_INFO_SEND = []
FLD_GET_IM_INFO_REC = [
    MessageField("address", 3, Address),
    MessageField("cat", 1, DeviceCategory),
    MessageField("subcat", 1, int),
    MessageField("firmware", 1, int),
    MessageField("ack", 1, AckNak),
]

# Send All-Link Command 0x61
FLD_SEND_ALL_LINK_CMD = [
    MessageField("group", 1, int),
    MessageField("cmd1", 1, int),
    MessageField("cmd2", 1, int),
]
FLD_SEND_ALL_LINK_CMD_ACK = FLD_SEND_ALL_LINK_CMD.copy()
FLD_SEND_ALL_LINK_CMD_ACK.append(MessageField("ack", 1, AckNak))

# Send INSTEON Standard or Extended Message 0x62
FLD_STD_SEND = [
    MessageField("address", 3, Address),
    MessageField("flags", 1, MessageFlags),
    MessageField("cmd1", 1, int),
    MessageField("cmd2", 1, int),
]
FLD_EXT_SEND = FLD_STD_SEND.copy()
FLD_EXT_SEND.append(MessageField("user_data", 14, UserData))

# Send INSTEON Standard or Extended Message ACK/NAK 0x62
FLD_STD_SEND_ACK = FLD_STD_SEND.copy()
FLD_STD_SEND_ACK.append(MessageField("ack", 1, AckNak))
FLD_EXT_SEND_ACK = FLD_EXT_SEND.copy()
FLD_EXT_SEND_ACK.append(MessageField("ack", 1, AckNak))

# Send X10 0x63 (SEE X10 Received 0x52)

# Start All-Linking 0x64
FLD_START_ALL_LINKING = [
    MessageField("link_mode", 1, AllLinkMode),
    MessageField("group", 1, int),
]
FLD_START_ALL_LINKING_ACK = FLD_START_ALL_LINKING.copy()
FLD_START_ALL_LINKING_ACK.append(MessageField("ack", 1, AckNak))

# Cancel All-Linking 0x65
FLD_CANCEL_ALL_LINKING = []
FLD_CANCEL_ALL_LINKING_ACK = [MessageField("ack", 1, AckNak)]

# Set Host Device Category 0x66
FLD_SET_HOST_DEVICE_CATEGORY = [
    MessageField("cat", 1, DeviceCategory),
    MessageField("subcat", 1, int),
    MessageField("firmware", 1, int),
]
FLD_SET_HOST_DEVICE_CATEGORY_ACK = FLD_SET_HOST_DEVICE_CATEGORY.copy()
FLD_SET_HOST_DEVICE_CATEGORY_ACK.append(MessageField("ack", 1, AckNak))

# Reset the IM 0x67
FLD_RESET_IM = []
FLD_RESET_IM_ACK = [MessageField("ack", 1, AckNak)]

# Set INSTEON ACK/NAK Message Byte 0x68/0x70
FLD_SET_ACK_NAK_BYTE = [MessageField("cmd2", 1, int)]
FLD_SET_ACK_NAK_BYTE_ACK = FLD_SET_ACK_NAK_BYTE.copy()
FLD_SET_ACK_NAK_BYTE_ACK.append(MessageField("ack", 1, AckNak))

# Get First ALL-Link Record 0x69
FLD_GET_FIRST_ALL_LINK_RECORD = []
FLD_GET_FIRST_ALL_LINK_RECORD_ACK = [MessageField("ack", 1, AckNak)]

# Get Next ALL-Link Record 0x6A
FLD_GET_NEXT_ALL_LINK_RECORD = []
FLD_GET_NEXT_ALL_LINK_RECORD_ACK = [MessageField("ack", 1, AckNak)]

# Set IM Configuration 0x6B
FLD_SET_IM_CONFIG = [MessageField("flags", 1, IMConfigurationFlags)]
FLD_SET_IM_CONFIG_ACK = FLD_SET_IM_CONFIG.copy()
FLD_SET_IM_CONFIG_ACK.append(MessageField("ack", 1, AckNak))

# Get All-Link Record for Sender 0x6C
FLD_GET_ALL_LINK_REC_FOR_SENDER = []
FLD_GET_ALL_LINK_REC_FOR_SENDER_ACK = [MessageField("ack", 1, AckNak)]

# LED On/Off 0x6D / 0x6E
FLD_LED_ON_OFF = []
FLD_LED_ON_OFF_ACK = [MessageField("ack", 1, AckNak)]

# Manage All-Link Record 0x6F
FLD_MANAGE_ALL_LINK_RECORD = [
    MessageField("action", 1, ManageAllLinkRecordAction),
    MessageField("flags", 1, AllLinkRecordFlags),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
    MessageField("data1", 1, int),
    MessageField("data2", 1, int),
    MessageField("data3", 1, int),
]
FLD_MANAGE_ALL_LINK_RECORD_ACK = FLD_MANAGE_ALL_LINK_RECORD.copy()
FLD_MANAGE_ALL_LINK_RECORD_ACK.append(MessageField("ack", 1, AckNak))

# Set INSTEON NAK Message Byte see 0x68
# Set INSTEON ACK Message Two Bytes 0x71
FLD_SET_ACK_TWO_BYTES = [MessageField("cmd1", 1, int), MessageField("cmd2", 1, int)]
FLD_SET_ACK_TWO_BYTES_ACK = FLD_SET_ACK_TWO_BYTES.copy()
FLD_SET_ACK_TWO_BYTES_ACK.append(MessageField("ack", 1, AckNak))

# RF Sleep 0x72
FLD_RF_SLEEP = []
FLD_RF_SLEEP_ACK = [MessageField("ack", 1, AckNak)]

# Get IM Configuration 0x73
FLD_GET_IM_CONFIG = []
FLD_GET_IM_CONFIG_ACK = [
    MessageField("flags", 1, IMConfigurationFlags),
    MessageField("spare1", 1, int),
    MessageField("spare2", 1, int),
    MessageField("ack", 1, AckNak),
]

# Read EEPROM 0x75
FLD_READ_EEPROM = [
    MessageField("mem_hi", 1, int),
    MessageField("mem_low", 1, int),
]
FLD_READ_EEPROM_ACK = FLD_READ_EEPROM.copy()
FLD_READ_EEPROM_ACK.append(MessageField("ack", 1, AckNak))

# Write to EEPROM 0x76
FLD_WRITE_EEPROM = [
    MessageField("mem_hi", 1, int),
    MessageField("mem_low", 1, int),
    MessageField("flags", 1, AllLinkRecordFlags),
    MessageField("group", 1, int),
    MessageField("target", 3, Address),
    MessageField("data1", 1, int),
    MessageField("data2", 1, int),
    MessageField("data3", 1, int),
]
FLD_WRITE_EEPROM_ACK = FLD_WRITE_EEPROM.copy()
FLD_WRITE_EEPROM_ACK.append(MessageField("ack", 1, AckNak))

INBOUND_MSG_DEF = {}
INBOUND_MSG_DEF[MessageId.STANDARD_RECEIVED] = MessageDefinition(
    MessageId.STANDARD_RECEIVED, FLD_STD_REC
)
INBOUND_MSG_DEF[MessageId.EXTENDED_RECEIVED] = MessageDefinition(
    MessageId.EXTENDED_RECEIVED, FLD_EXT_REC
)
INBOUND_MSG_DEF[MessageId.X10_RECEIVED] = MessageDefinition(
    MessageId.X10_RECEIVED, FLD_X10_SEND_REC
)
INBOUND_MSG_DEF[MessageId.ALL_LINKING_COMPLETED] = MessageDefinition(
    MessageId.ALL_LINKING_COMPLETED, FLD_ALL_LINK_COMPLETE
)
INBOUND_MSG_DEF[MessageId.BUTTON_EVENT_REPORT] = MessageDefinition(
    MessageId.BUTTON_EVENT_REPORT, FLD_IM_BTN_EVENT_REPORT_REC
)
INBOUND_MSG_DEF[MessageId.USER_RESET_DETECTED] = MessageDefinition(
    MessageId.USER_RESET_DETECTED, FLD_USER_RESET_REC
)
INBOUND_MSG_DEF[MessageId.ALL_LINK_CLEANUP_FAILURE_REPORT] = MessageDefinition(
    MessageId.ALL_LINK_CLEANUP_FAILURE_REPORT, FLD_ALL_LINK_CLEANUP_FAILURE
)
INBOUND_MSG_DEF[MessageId.ALL_LINK_RECORD_RESPONSE] = MessageDefinition(
    MessageId.ALL_LINK_RECORD_RESPONSE, FLD_ALL_LINK_RECORD_RESP
)
INBOUND_MSG_DEF[MessageId.ALL_LINK_CLEANUP_STATUS_REPORT] = MessageDefinition(
    MessageId.ALL_LINK_CLEANUP_STATUS_REPORT, FLD_ALL_LINK_CLEANUP_REPORT_ACK
)
INBOUND_MSG_DEF[MessageId.READ_EEPROM_RESPONSE] = MessageDefinition(
    MessageId.READ_EEPROM_RESPONSE, FLD_READ_EEPROM_RESPONSE
)
INBOUND_MSG_DEF[MessageId.GET_IM_INFO] = MessageDefinition(
    MessageId.GET_IM_INFO, FLD_GET_IM_INFO_REC
)
INBOUND_MSG_DEF[MessageId.SEND_ALL_LINK_COMMAND] = MessageDefinition(
    MessageId.SEND_ALL_LINK_COMMAND, FLD_SEND_ALL_LINK_CMD_ACK
)
INBOUND_MSG_DEF[MessageId.SEND_STANDARD] = MessageDefinition(
    MessageId.SEND_STANDARD, FLD_STD_SEND_ACK
)
INBOUND_MSG_DEF[MessageId.X10_SEND] = MessageDefinition(
    MessageId.X10_SEND, FLD_X10_SEND_REC_ACK
)
INBOUND_MSG_DEF[MessageId.START_ALL_LINKING] = MessageDefinition(
    MessageId.START_ALL_LINKING, FLD_START_ALL_LINKING_ACK
)
INBOUND_MSG_DEF[MessageId.CANCEL_ALL_LINKING] = MessageDefinition(
    MessageId.CANCEL_ALL_LINKING, FLD_CANCEL_ALL_LINKING_ACK
)
INBOUND_MSG_DEF[MessageId.SET_HOST_DEVICE_CATEGORY] = MessageDefinition(
    MessageId.SET_HOST_DEVICE_CATEGORY, FLD_SET_HOST_DEVICE_CATEGORY_ACK
)
INBOUND_MSG_DEF[MessageId.RESET_IM] = MessageDefinition(
    MessageId.RESET_IM, FLD_RESET_IM_ACK
)
INBOUND_MSG_DEF[MessageId.SET_ACK_MESSAGE_BYTE] = MessageDefinition(
    MessageId.SET_ACK_MESSAGE_BYTE, FLD_SET_ACK_NAK_BYTE_ACK
)
INBOUND_MSG_DEF[MessageId.GET_FIRST_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.GET_FIRST_ALL_LINK_RECORD, FLD_GET_FIRST_ALL_LINK_RECORD_ACK
)
INBOUND_MSG_DEF[MessageId.GET_NEXT_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.GET_NEXT_ALL_LINK_RECORD, FLD_GET_NEXT_ALL_LINK_RECORD_ACK
)
INBOUND_MSG_DEF[MessageId.SET_IM_CONFIGURATION] = MessageDefinition(
    MessageId.SET_IM_CONFIGURATION, FLD_SET_IM_CONFIG_ACK
)
INBOUND_MSG_DEF[MessageId.GET_ALL_LINK_RECORD_FOR_SENDER] = MessageDefinition(
    MessageId.GET_ALL_LINK_RECORD_FOR_SENDER, FLD_GET_ALL_LINK_REC_FOR_SENDER_ACK
)
INBOUND_MSG_DEF[MessageId.LED_ON] = MessageDefinition(
    MessageId.LED_ON, FLD_LED_ON_OFF_ACK
)
INBOUND_MSG_DEF[MessageId.LED_OFF] = MessageDefinition(
    MessageId.LED_OFF, FLD_LED_ON_OFF_ACK
)
INBOUND_MSG_DEF[MessageId.MANAGE_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.MANAGE_ALL_LINK_RECORD, FLD_MANAGE_ALL_LINK_RECORD_ACK
)
INBOUND_MSG_DEF[MessageId.SET_NAK_MESSAGE_BYTE] = MessageDefinition(
    MessageId.SET_NAK_MESSAGE_BYTE, FLD_SET_ACK_NAK_BYTE_ACK
)
INBOUND_MSG_DEF[MessageId.SET_ACK_MESSAGE_TWO_BYTES] = MessageDefinition(
    MessageId.SET_ACK_MESSAGE_TWO_BYTES, FLD_SET_ACK_TWO_BYTES_ACK
)
INBOUND_MSG_DEF[MessageId.RF_SLEEP] = MessageDefinition(
    MessageId.RF_SLEEP, FLD_RF_SLEEP_ACK
)
INBOUND_MSG_DEF[MessageId.GET_IM_CONFIGURATION] = MessageDefinition(
    MessageId.GET_IM_CONFIGURATION, FLD_GET_IM_CONFIG_ACK
)
INBOUND_MSG_DEF[MessageId.READ_EEPROM] = MessageDefinition(
    MessageId.READ_EEPROM, FLD_READ_EEPROM_ACK
)
INBOUND_MSG_DEF[MessageId.WRITE_EEPROM] = MessageDefinition(
    MessageId.WRITE_EEPROM, FLD_WRITE_EEPROM_ACK
)

OUTBOUND_MSG_DEF = {}
OUTBOUND_MSG_DEF[MessageId.GET_IM_INFO] = MessageDefinition(
    MessageId.GET_IM_INFO, FLD_GET_IM_INFO_SEND
)
OUTBOUND_MSG_DEF[MessageId.SEND_ALL_LINK_COMMAND] = MessageDefinition(
    MessageId.SEND_ALL_LINK_COMMAND, FLD_SEND_ALL_LINK_CMD
)
OUTBOUND_MSG_DEF[MessageId.SEND_STANDARD] = MessageDefinition(
    MessageId.SEND_STANDARD, FLD_STD_SEND
)
OUTBOUND_MSG_DEF[MessageId.X10_SEND] = MessageDefinition(
    MessageId.X10_SEND, FLD_X10_SEND_REC
)
OUTBOUND_MSG_DEF[MessageId.START_ALL_LINKING] = MessageDefinition(
    MessageId.START_ALL_LINKING, FLD_START_ALL_LINKING
)
OUTBOUND_MSG_DEF[MessageId.CANCEL_ALL_LINKING] = MessageDefinition(
    MessageId.CANCEL_ALL_LINKING, FLD_CANCEL_ALL_LINKING
)
OUTBOUND_MSG_DEF[MessageId.SET_HOST_DEVICE_CATEGORY] = MessageDefinition(
    MessageId.SET_HOST_DEVICE_CATEGORY, FLD_SET_HOST_DEVICE_CATEGORY
)
OUTBOUND_MSG_DEF[MessageId.RESET_IM] = MessageDefinition(
    MessageId.RESET_IM, FLD_RESET_IM
)
OUTBOUND_MSG_DEF[MessageId.SET_ACK_MESSAGE_BYTE] = MessageDefinition(
    MessageId.SET_ACK_MESSAGE_BYTE, FLD_SET_ACK_NAK_BYTE
)
OUTBOUND_MSG_DEF[MessageId.GET_FIRST_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.GET_FIRST_ALL_LINK_RECORD, FLD_GET_FIRST_ALL_LINK_RECORD
)
OUTBOUND_MSG_DEF[MessageId.GET_NEXT_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.GET_NEXT_ALL_LINK_RECORD, FLD_GET_NEXT_ALL_LINK_RECORD
)
OUTBOUND_MSG_DEF[MessageId.SET_IM_CONFIGURATION] = MessageDefinition(
    MessageId.SET_IM_CONFIGURATION, FLD_SET_IM_CONFIG
)
OUTBOUND_MSG_DEF[MessageId.GET_ALL_LINK_RECORD_FOR_SENDER] = MessageDefinition(
    MessageId.GET_ALL_LINK_RECORD_FOR_SENDER, FLD_GET_ALL_LINK_REC_FOR_SENDER
)
OUTBOUND_MSG_DEF[MessageId.LED_ON] = MessageDefinition(MessageId.LED_ON, FLD_LED_ON_OFF)
OUTBOUND_MSG_DEF[MessageId.LED_OFF] = MessageDefinition(
    MessageId.LED_OFF, FLD_LED_ON_OFF
)
OUTBOUND_MSG_DEF[MessageId.MANAGE_ALL_LINK_RECORD] = MessageDefinition(
    MessageId.MANAGE_ALL_LINK_RECORD, FLD_MANAGE_ALL_LINK_RECORD
)
OUTBOUND_MSG_DEF[MessageId.SET_NAK_MESSAGE_BYTE] = MessageDefinition(
    MessageId.SET_NAK_MESSAGE_BYTE, FLD_SET_ACK_NAK_BYTE
)
OUTBOUND_MSG_DEF[MessageId.SET_ACK_MESSAGE_TWO_BYTES] = MessageDefinition(
    MessageId.SET_ACK_MESSAGE_TWO_BYTES, FLD_SET_ACK_TWO_BYTES
)
OUTBOUND_MSG_DEF[MessageId.RF_SLEEP] = MessageDefinition(
    MessageId.RF_SLEEP, FLD_RF_SLEEP
)
OUTBOUND_MSG_DEF[MessageId.GET_IM_CONFIGURATION] = MessageDefinition(
    MessageId.GET_IM_CONFIGURATION, FLD_GET_IM_CONFIG
)

OUTBOUND_MSG_DEF[MessageId.READ_EEPROM] = MessageDefinition(
    MessageId.READ_EEPROM, FLD_READ_EEPROM
)

OUTBOUND_MSG_DEF[MessageId.WRITE_EEPROM] = MessageDefinition(
    MessageId.WRITE_EEPROM, FLD_WRITE_EEPROM
)



"""Base class for all subscribable classes."""
import logging
from abc import ABC
from typing import Callable

from .utils import publish_topic, subscribe_topic, unsubscribe_topic

_LOGGER = logging.getLogger(__name__)


class SubscriberBase(ABC):
    """Event class to manage triggering of events."""

    def __init__(self, subscriber_topic):
        """Init the Event class."""
        self._subscriber_topic = subscriber_topic
        self._subscribers = []

    @property
    def topic(self):
        """Return the subscriber topic."""
        return self._subscriber_topic

    def subscribe(self, callback: Callable, force_strong_ref=False):
        """Subscribe to the event."""
        if force_strong_ref and callback not in self._subscribers:
            _LOGGER.debug("Adding subscriber to persistant list")
            self._subscribers.append(callback)
        subscribe_topic(callback, self._subscriber_topic, _LOGGER)

    def unsubscribe(self, callback):
        """Unsubscribe to the event."""
        unsubscribe_topic(callback, self._subscriber_topic)

    def _call_subscribers(self, **kwargs):
        """Call subscribers to the event."""
        publish_topic(self._subscriber_topic, **kwargs)



"""Insteon Powerline Modem (PLM)."""

from ..aldb.modem_aldb import ModemALDB
from .modem_base import ModemBase


class PLM(ModemBase):
    """Insteon PLM class."""

    def __init__(
        self,
        address="000000",
        cat=0x03,
        subcat=0x00,
        firmware=0x00,
        description="",
        model="",
    ):
        """Init the Modem class."""
        super().__init__(address, cat, subcat, firmware, description, model)
        if subcat in [0x05, 0x06, 0x0B, 0x10, 0x13, 0x14]:
            mem_addr = 0x3FFF
        else:
            mem_addr = 0x1FFF
        self._aldb = ModemALDB(self._address, mem_addr=mem_addr)



"""Message Flags data type."""

import binascii
import logging

from ..constants import MessageFlagType
from ..utils import test_values_eq

_LOGGER = logging.getLogger(__name__)
_EXTENDED_MESSAGE = 0x10


class MessageFlags:
    """Message Flags class use in Standard and Extended messages."""

    def __init__(self, flags=0x00):
        """Init the MessageFlags class."""
        self._type = None
        self._extended = None
        self._hops_left = None
        self._max_hops = None

        if flags is not None:
            self._set_properties(flags)

    def __repr__(self):
        """Representation of the message flags."""
        int_self = int.from_bytes(bytes(self), byteorder="big")
        return f"0x{int_self:02x}"

    def __str__(self):
        """Return a hexadecimal representation of the message flags."""
        start = "{"
        end = "}"
        val = f"{start}'message_type': '{str(self._type)}', 'extended': {self._extended}, 'hops_left': {self._hops_left}, 'max_hops': {self._max_hops}{end}"
        return val

    def __bytes__(self):
        """Return a byte representation of the message flags."""
        message_type = (self._type.value << 5) if self._type else 0
        extended_bit = (1 << 4) if self._extended else 0
        hops_left = (self._hops_left << 2) if self._hops_left else 0
        hops_max = self._max_hops if self._max_hops else 0
        flag_byte = message_type | extended_bit | hops_left | hops_max
        return bytes([flag_byte])

    def __eq__(self, other):
        """Test for equality.

        Test only message type and extended bit. Ignore hops remaining and
        max hops.
        """
        if not isinstance(other, MessageFlags):
            return False

        match = True
        match = match & test_values_eq(self.message_type, other.message_type)
        match = match & test_values_eq(self.extended, other.extended)
        return match

    def __ne__(self, other):
        """Test for inequality.

        Test only message type and extended bit. Ignore hops remaining and
        max hops.
        """
        return not self.__eq__(other)

    def __hash__(self):
        """Represent the MessageFlags object as a hash."""
        return hash(bytes(self))

    @property
    def message_type(self) -> MessageFlagType:
        """Return the message type."""
        return self._type

    @message_type.setter
    def message_type(self, val: MessageFlagType):
        """Set the message type."""
        if isinstance(val, MessageFlagType) or val is None:
            self._type = val
        elif val in range(0, 8):
            self._type = MessageFlagType(val)
        else:
            raise ValueError("message_type property must be a MessageFlagType.")

    @property
    def hops_left(self) -> int:
        """Return the number of hops left in message the trasport."""
        return self._hops_left

    @hops_left.setter
    def hops_left(self, val: int):
        """Set the number of hops left for this message."""
        if isinstance(val, int):
            val_int = val
        elif isinstance(val, bytes):
            val_int = int.from_bytes(val, byteorder="big")
        elif val is None:
            self._hops_left = val
            return
        else:
            raise ValueError("hops_left property must be 0-3 or None")
        self._hops_left = min(3, max(val_int, 0))

    @property
    def max_hops(self) -> int:
        """Return the maximum number of hops allowed for this message."""
        return self._max_hops

    @max_hops.setter
    def max_hops(self, val: int):
        """Set the maximum number of hops allowed for this message."""
        if isinstance(val, int):
            val_int = val
        elif isinstance(val, bytes):
            val_int = int.from_bytes(val, byteorder="big")
        elif val is None:
            self._max_hops = val
            return
        else:
            raise ValueError("max_hops property must be 0-3 or None")
        self._max_hops = min(3, max(val_int, 0))

    @property
    def extended(self) -> MessageFlagType:
        """Return the extended flag."""
        return self._extended

    @extended.setter
    def extended(self, val: bool):
        """Set the extended flag."""
        if val in [None, True, False]:
            self._extended = val
        else:
            try:
                self._extended = bool(val)
            except TypeError as ex:
                raise ValueError(
                    "extended property must be True, False or None."
                ) from ex

    @property
    def is_broadcast(self) -> bool:
        """Test if the message is a broadcast message type."""
        return self._type == MessageFlagType.BROADCAST

    @property
    def is_direct(self) -> bool:
        """Test if the message is a direct message type."""
        direct = self._type == MessageFlagType.DIRECT
        if self.is_direct_ack or self.is_direct_nak:
            direct = True
        return direct

    @property
    def is_direct_ack(self) -> bool:
        """Test if the message is a direct ACK message type."""
        return self._type == MessageFlagType.DIRECT_ACK

    @property
    def is_direct_nak(self) -> bool:
        """Test if the message is a direct NAK message type."""
        return self._type == MessageFlagType.DIRECT_NAK

    @property
    def is_all_link_broadcast(self) -> bool:
        """Test if the message is an ALl-Link broadcast message type."""
        return self._type == MessageFlagType.ALL_LINK_BROADCAST

    @property
    def is_all_link_cleanup(self) -> bool:
        """Test if the message is a All-Link cleanup message type."""
        return self._type == MessageFlagType.ALL_LINK_CLEANUP

    @property
    def is_all_link_cleanup_ack(self) -> bool:
        """Test if the message is a All-LInk cleanup ACK message type."""
        return self._type == MessageFlagType.ALL_LINK_CLEANUP_ACK

    @property
    def is_all_link_cleanup_nak(self) -> bool:
        """Test if the message is a All-Link cleanup NAK message type."""
        return self._type == MessageFlagType.ALL_LINK_CLEANUP_NAK

    @property
    def is_extended(self) -> bool:
        """Test if the message is an extended message type."""
        return self._extended == 1

    def _normalize(self, flags) -> bytes:
        """Take any format of flags and turn it into a hex string."""
        norm = None
        if isinstance(flags, MessageFlags):
            norm = bytes(flags)
        elif isinstance(flags, bytearray):
            norm = binascii.hexlify(flags)
        elif isinstance(flags, int):
            norm = bytes([flags])
        elif isinstance(flags, bytes):
            norm = binascii.hexlify(flags)
        elif isinstance(flags, str):
            flags = flags[0:2]
            norm = binascii.hexlify(binascii.unhexlify(flags.lower()))
        elif flags is None:
            norm = None
        else:
            _LOGGER.warning("MessageFlags with unknown type %s: %r", type(flags), flags)
        return norm

    def _set_properties(self, flags):
        """Set the properties of the message flags based on a byte input."""
        flag_byte = self._normalize(flags)

        if flag_byte is not None:
            self._type = MessageFlagType((flag_byte[0] & 0xE0) >> 5)
            self._extended = (flag_byte[0] & _EXTENDED_MESSAGE) >> 4
            self._hops_left = (flag_byte[0] & 0x0C) >> 2
            self._max_hops = flag_byte[0] & 0x03
        else:
            self._type = None
            self._extended = None
            self._hops_left = None
            self._max_hops = None

    @classmethod
    def create(
        cls,
        message_flag_type: MessageFlagType,
        extended: bool = False,
        hops_left: int = 0,
        max_hops: int = 0,
    ):
        """Create message flags.

        message_flag_type: MessageFlagType 0 to 7:
            DIRECT = 0
            DIRECT_ACK = 1
            ALL_LINK_CLEANUP = 2
            ALL_LINK_CLEANUP_ACK = 3
            BROADCAST = 4
            DIRECT_NAK = 5
            ALL_LINK_BROADCAST = 6
            ALL_LINK_CLEANUP_NAK = 7
        extended: True for extended, False for standard
        hops_left: int  0 - 3
        max_hops:  int  0 - 3
        """
        flags = cls(0x00)
        flags.message_type = message_flag_type
        flags.extended = extended
        flags.hops_left = hops_left
        flags.max_hops = max_hops
        return flags



"""Manage the inbound messages that trigger a variable state's on level."""
from datetime import datetime, timedelta

from ..address import Address
from ..handlers.from_device.off import OffInbound
from ..handlers.from_device.off_all_link_cleanup import OffAllLinkCleanupInbound
from ..handlers.from_device.off_fast import OffFastInbound
from ..handlers.from_device.off_fast_all_link_cleanup import (
    OffFastAllLinkCleanupInbound,
)
from ..handlers.from_device.on_fast import OnFastInbound
from ..handlers.from_device.on_fast_all_link_cleanup import OnFastAllLinkCleanupInbound
from ..handlers.from_device.on_level import OnLevelInbound
from ..handlers.from_device.on_level_all_link_cleanup import OnAllLinkCleanupInbound
from ..subscriber_base import SubscriberBase

TIMEOUT_CLEANUP = timedelta(seconds=25)


class OnLevelManager:
    """Manage the inbound messages that trigger a variable state's on level.

    The manager is used in all variable state devices. These include:
        - Category 0 devices such as Mini-Remotes
        - All category 1 dimmable devices
        - All category 2 switch devices
        - etc

    This manager handles inbound broadcast messages that identify a state change of
    a device. It also handles inbound clean up messages that follow the broadcast
    message. Finally, it is responsible to deduplicate messages to ensure multiple
    broadcast and cleanup messages to the same group for the same state change
    only trigger once.
    """

    class Subscriber(SubscriberBase):
        """Internal class to trigger notification of events or state values."""

        def call_subscribers(self, on_level):
            """Call subscribers to this manager for the event type."""
            self._call_subscribers(on_level=on_level)

    def __init__(self, address, group, default_on_level=0xFF):
        """Init the OnLevelManager class."""
        self._address = Address(address)
        self._group = int(group)
        self._default_on_level = default_on_level
        self._last_event = datetime(1, 1, 1, 1, 1, 1)
        self._last_event_type = None

        # Setup event managers that will manange the subscribers to specific events
        self._on = self.Subscriber(
            f"subscriber_{self._address.id}_on_{self._group}_broadcast"
        )
        self._off = self.Subscriber(
            f"subscriber_{self._address.id}_off_{self._group}_broadcast"
        )
        self._on_fast = self.Subscriber(
            f"subscriber_{self._address.id}_on_fast_{self._group}_broadcast"
        )
        self._off_fast = self.Subscriber(
            f"subscriber_{self._address.id}_off_fast_{self._group}_broadcast"
        )

        # Register the handlers to listen to
        self._on_handler = OnLevelInbound(self._address, self._group)
        self._off_handler = OffInbound(self._address, self._group)
        self._on_fast_handler = OnFastInbound(self._address, self._group)
        self._off_fast_handler = OffFastInbound(self._address, self._group)
        self._on_cleanup_handler = OnAllLinkCleanupInbound(self._address, self._group)
        self._off_cleanup_handler = OffAllLinkCleanupInbound(self._address, self._group)
        self._on_fast_cleanup_handler = OnFastAllLinkCleanupInbound(
            self._address, self._group
        )
        self._off_fast_cleanup_handler = OffFastAllLinkCleanupInbound(
            self._address, self._group
        )

        # Subscribe to events
        self._on_handler.subscribe(self._on_event)
        self._off_handler.subscribe(self._off_event)
        self._on_fast_handler.subscribe(self._on_fast_event)
        self._off_fast_handler.subscribe(self._off_fast_event)
        self._on_cleanup_handler.subscribe(self._on_cleanup_event)
        self._off_cleanup_handler.subscribe(self._off_cleanup_event)
        self._on_fast_cleanup_handler.subscribe(self._on_fast_cleanup_event)
        self._off_fast_cleanup_handler.subscribe(self._off_fast_cleanup_event)

    def subscribe(self, callback, force_strong_ref=False):
        """Subscribe to all events (ON, OFF, ON FAST, OFF FAST)."""
        self._on.subscribe(callback, force_strong_ref=force_strong_ref)
        self._off.subscribe(callback, force_strong_ref=force_strong_ref)
        self._on_fast.subscribe(callback, force_strong_ref=force_strong_ref)
        self._off_fast.subscribe(callback, force_strong_ref=force_strong_ref)

    def subscribe_on(self, callback, force_strong_ref=False):
        """Subscribe to ON events."""
        self._on.subscribe(callback, force_strong_ref=force_strong_ref)

    def subscribe_off(self, callback, force_strong_ref=False):
        """Subscribe to OFF events."""
        self._off.subscribe(callback, force_strong_ref=force_strong_ref)

    def subscribe_on_fast(self, callback, force_strong_ref=False):
        """Subscribe to ON FAST events."""
        self._on_fast.subscribe(callback, force_strong_ref=force_strong_ref)

    def subscribe_off_fast(self, callback, force_strong_ref=False):
        """Subscribe to OFF FAST events."""
        self._off_fast.subscribe(callback, force_strong_ref=force_strong_ref)

    def _on_event(self, on_level):
        self._process_on_event(on_level=on_level)

    def _off_event(self, on_level):
        self._process_off_event()

    def _on_fast_event(self, on_level):
        self._process_on_fast_event(on_level=on_level)

    def _off_fast_event(self, on_level):
        self._process_off_fast_event()

    def _on_cleanup_event(self):
        self._process_on_event(on_level=self._default_on_level, is_cleanup=True)

    def _off_cleanup_event(self):
        self._process_off_event(is_cleanup=True)

    def _on_fast_cleanup_event(self):
        self._process_on_fast_event(on_level=self._default_on_level, is_cleanup=True)

    def _off_fast_cleanup_event(self):
        self._process_off_fast_event(is_cleanup=True)

    def _process_on_event(self, on_level, is_cleanup=False):
        if not is_cleanup or self._process_event("on"):
            self._on.call_subscribers(on_level=on_level)
            self._last_event = datetime.now()
            self._last_event_type = "on"

    def _process_off_event(self, is_cleanup=False):
        if not is_cleanup or self._process_event("off"):
            self._off.call_subscribers(on_level=0)
            self._last_event = datetime.now()
            self._last_event_type = "off"

    def _process_on_fast_event(self, on_level, is_cleanup=False):
        if not is_cleanup or self._process_event("on_fast"):
            self._on_fast.call_subscribers(on_level=on_level)
            self._last_event = datetime.now()
            self._last_event_type = "on_fast"

    def _process_off_fast_event(self, is_cleanup=False):
        if not is_cleanup or self._process_event("off_fast"):
            self._off_fast.call_subscribers(on_level=0)
            self._last_event = datetime.now()
            self._last_event_type = "off_fast"

    def _process_event(self, event_type):
        tdelta = datetime.now() - self._last_event
        if self._last_event_type == event_type and tdelta < TIMEOUT_CLEANUP:
            return False
        return True



"""Scenes menu for pyinsteon tools."""

from .. import devices
from ..constants import DeviceCategory
from ..managers.device_link_manager import DeviceLinkSchema
from ..managers.scene_manager import (
    async_add_or_update_scene,
    async_get_scene,
    async_get_scenes,
    async_save_scene_names,
    set_scene_name,
)
from .tools_base import ToolsBase

SHOW_ADVANCED = True


class ToolsScenes(ToolsBase):
    """Command class to test interactivity."""

    async def do_list_scenes(self):
        """List all scene."""
        self._log_stdout("Number   Name")
        self._log_stdout(
            "------   -------------------------------------------------------"
        )
        scenes = await async_get_scenes()
        for scene, data in scenes.items():
            self._log_stdout(f"{scene:5}    {data['name']}")

    async def do_print_scene(
        self, scene: int = None, workdir: str = None, log_stdout=None, background=False
    ):
        """Show the details of a scene.

        Usage:
            print_scene scene

        scene: Scene number
        """
        if workdir is None:
            workdir = self.workdir if self.workdir else "."

        self.workdir = workdir

        if not self.workdir:
            if background:
                log_stdout("A value for the working directory is required.")
                return
            self.workdir = await self._get_workdir()

        if not self.workdir:
            return

        scene_list = list(await async_get_scenes(work_dir=self.workdir))

        scene = await self._ensure_int(
            value=scene,
            values=scene_list,
            name="Scene number",
            ask_value=not background,
            log_stdout=log_stdout,
        )
        if not scene:
            return

        scene_def = await async_get_scene(scene, work_dir=self.workdir)
        log_stdout(f"Scene {scene}: {scene_def['name']}")
        log_stdout("")
        log_stdout("Device     Data 1   Data 2   Data 3")
        log_stdout("--------   ------   ------   ------")
        for device, info in scene_def["devices"].items():
            data1 = f"{info.data1:5}" if info.data1 is not None else "     "
            data2 = f"{info.data2:5}" if info.data2 is not None else "     "
            data3 = f"{info.data3:5}" if info.data3 is not None else "     "
            log_stdout(f"{str(device)}   {data1}   {data2}   {data3}")

    async def do_save_scenes(self, workdir=None, log_stdout=None, background=False):
        """Save the scene names to the working directory.

        Usage:
            save_devices [--background | -b] workdir

        workdir: Directory where the saved device file is located

        This does not write the scenes to the devices. Use `write_scenes` to write the scene data
        to the devices.
        """
        if workdir is None:
            workdir = self.workdir if self.workdir else "."

        self.workdir = workdir

        if not self.workdir:
            if background:
                log_stdout("A value for the working directory is required.")
                return
            self.workdir = await self._get_workdir()

        if not self.workdir:
            return

        await async_save_scene_names(work_dir=self.workdir)

    async def do_set_scene_name(
        self, scene=None, name=None, work_dir=None, log_stdout=None, background=False
    ):
        """Set the scene name."""
        scene_list = list(await async_get_scenes(work_dir=self.workdir))

        scene = await self._ensure_int(
            value=scene,
            values=scene_list,
            name="Scene number",
            ask_value=not background,
            log_stdout=log_stdout,
        )

        name = await self._ensure_string(
            value=name,
            values=None,
            name="Scene name",
            ask_value=not background,
            log_stdout=log_stdout,
        )

        if not scene:
            return

        set_scene_name(scene_num=scene, name=name)

    async def do_add_scene(self):
        """Add a new scene.

        Usage:
            add_scene

        Note: This cannot be run in the background. It is an interactive method which will provide prompts for scene information.
        """

        scene_info = {}
        scene = None
        scene_list = list(async_get_scenes())
        while scene is None or scene in scene_list:
            scene = self._ensure_int(
                value=scene,
                values=None,
                name="Scene number",
                ask_value=True,
                log_stdout=self._log_stdout,
            )
            if scene in scene_list:
                self._log_stdout(f"Scene number {scene} already exists.")
            if not scene:
                return

        default_name = f"Scene number {scene}"
        name = await self._ensure_string(
            value=None,
            values=None,
            name="Scene Name",
            ask_value=True,
            log_stdout=self._log_stdout,
            default=default_name,
        )
        scene_info["name"] = name
        scene_devices = []
        got_addr = True
        while got_addr:
            addr = await self._ensure_address(
                address=None,
                name="Device Address",
                ask_value=True,
                log_stdout=self._log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addr:
                got_addr = False
                continue

            device = devices[addr]
            got_data = True
            if device.cat == DeviceCategory.DIMMABLE_LIGHTING_CONTROL:
                data1_values = range(0, 255)
                data2_values = range(0, 255)
            else:
                data1_values = [0, 255]
                data2_values = []
            data3_values = list(device.groups)

            device_data = []
            while got_data:
                data1 = await self._ensure_int(
                    value=None,
                    values=data1_values,
                    name="Data 1",
                    ask_value=True,
                    log_stdout=self._log_stdout,
                    default=255,
                )
                if data2_values:
                    data2 = await self._ensure_int(
                        value=None,
                        values=data2_values,
                        name="Data 2",
                        ask_value=True,
                        log_stdout=self._log_stdout,
                        default=28,
                    )
                else:
                    data2 = 0

                data3 = await self._ensure_int(
                    value=None,
                    values=data3_values,
                    name="Data 3",
                    ask_value=True,
                    log_stdout=self._log_stdout,
                    default=1,
                )
                data = {
                    "address": device.address.id,
                    "data1": data1,
                    "data2": data2,
                    "data3": data3,
                }
                device_data.append(data)
            scene_devices = DeviceLinkSchema(device_data)
        if not self.workdir:
            self.workdir = await self._get_workdir()
            if not self.workdir:
                self._log_stdout(
                    "A working directory is required to save the scene name."
                )
        if not scene_devices:
            self._log_stdout("At least one device is required")
            return
        await async_add_or_update_scene(
            scene_num=scene, links=scene_devices, name=name, work_dir=self.workdir
        )



"""IOLink relay mode property."""
from ..constants import RelayMode
from . import RELAY_MODE, get_usable_value
from .derived_property import DerivedProperty


def _calc_relay_mode(
    momentary_mode_on_value,
    momentary_follow_sense_value,
    momentary_on_off_trigger_value,
):
    """Calculate the relay mode."""
    if not momentary_mode_on_value:
        return RelayMode.LATCHING
    if momentary_follow_sense_value:
        return RelayMode.MOMENTARY_C
    if momentary_on_off_trigger_value:
        return RelayMode.MOMENTARY_B
    return RelayMode.MOMENTARY_A


class RelayModeProperty(DerivedProperty):
    """IOLink relay mode property."""

    def __init__(
        self,
        address,
        name,
        momentary_mode_on_prop,
        momentary_follow_sense_prop,
        momentary_on_off_trigger_prop,
    ):
        """Init the ExtendedProperty class."""
        super().__init__(address, RELAY_MODE, RelayMode, False, False)

        self._momentary_mode_on_prop = momentary_mode_on_prop
        self._momentary_follow_sense_prop = momentary_follow_sense_prop
        self._momentary_on_off_trigger_prop = momentary_on_off_trigger_prop

    @property
    def value(self):
        """Return the relay mode based on the underlying properties."""
        on_prop = self._momentary_mode_on_prop
        f_prop = self._momentary_follow_sense_prop
        oo_prop = self._momentary_on_off_trigger_prop

        return _calc_relay_mode(on_prop.value, f_prop.value, oo_prop.value)

    @property
    def new_value(self):
        """Return the new_value property."""
        on_val = get_usable_value(self._momentary_mode_on_prop)
        follow_val = get_usable_value(self._momentary_follow_sense_prop)
        on_off_val = get_usable_value(self._momentary_on_off_trigger_prop)

        new_value = _calc_relay_mode(on_val, follow_val, on_off_val)
        if new_value == self.value:
            return None
        return new_value

    @new_value.setter
    def new_value(self, value):
        """Set the new_value property."""

        def set_relay_mode(momentary_mode_on, momentary_follow_sense, momentary_on_off):
            """Set the values of the underlying properties."""
            self._momentary_mode_on_prop.new_value = momentary_mode_on
            self._momentary_follow_sense_prop.new_value = momentary_follow_sense
            self._momentary_on_off_trigger_prop.new_value = momentary_on_off

        if value is None:
            set_relay_mode(None, None, None)
            return

        relay_mode = RelayMode(int(value))
        if relay_mode == self.value:
            set_relay_mode(None, None, None)
            return

        if relay_mode == RelayMode.LATCHING:
            set_relay_mode(False, False, False)
        elif relay_mode == RelayMode.MOMENTARY_A:
            set_relay_mode(True, False, False)
        elif relay_mode == RelayMode.MOMENTARY_B:
            set_relay_mode(True, False, True)
        elif relay_mode == RelayMode.MOMENTARY_C:
            set_relay_mode(True, True, False)

    @property
    def is_dirty(self):
        """Return the change status of the property."""
        return (
            self._momentary_follow_sense_prop.is_dirty
            or self._momentary_mode_on_prop.is_dirty
            or self._momentary_on_off_trigger_prop.is_dirty
        )

    @property
    def is_loaded(self):
        """Return the change status of the property."""
        return (
            self._momentary_follow_sense_prop.is_loaded
            and self._momentary_mode_on_prop.is_loaded
            and self._momentary_on_off_trigger_prop.is_loaded
        )
